/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: checkrom.c,v $
 * Revision 1.2  1994/10/06  16:30:05  ecd
 * changed char to unsigned
 *
 * Revision 1.1  1994/10/01  10:12:53  ecd
 * Initial revision
 *
 *
 * $Id: checkrom.c,v 1.2 1994/10/06 16:30:05 ecd Exp ecd $
 */

#include <stdio.h>
#include <stdlib.h>

#include "global.h"

#define ROM_SIZE	0x80000
unsigned char rom[ROM_SIZE];

unsigned short rom_crc, crc;

#define calc_crc(n) (crc = ((crc >> 4) ^ (((crc ^ n) & 0xf) * 0x1081)))

int
#ifdef __FunctionProto__
main(int argc, char **argv)
#else
main(argc, argv)
int argc;
char **argv;
#endif
{
  FILE *fp;
  unsigned char version[7];
  int i, a, c, d, d0, d1, D0, D1;

  if (argc < 2) {
    fprintf(stderr, "usage: %s rom-file\n", argv[0]);
    exit (1);
  }
  if ((fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "%s: can\'t open %s\n", argv[0], argv[1]);
    exit (1);
  }
  if (fread(rom, 1, ROM_SIZE, fp) != ROM_SIZE) {
    fprintf(stderr, "%s: can\'t read %d bytes from %s\n",
            argv[0], ROM_SIZE, argv[1]);
    fclose(fp);
    exit (1);
  }
  fclose(fp);

  for (i = 0; i < 6; i++) {
    version[i] = rom[ROM_SIZE + 2 * i - 15] << 4;
    version[i] |= rom[ROM_SIZE + 2 * i - 16];
  }
  version[6] = '\0';

  rom_crc = 0;
  for (i = 0; i < 4; i++) {
    rom_crc <<= 4;
    rom_crc |= rom[ROM_SIZE - i - 1] & 0x0f;
  }
  printf("ROM Version %s CRC reads 0x%.4x\n", version, rom_crc);

  for (i = 0x100; i < 0x140; i++) {
    rom[i] = 0x0;
  }

  d = (ROM_SIZE / 0x80000);
  crc = 0x0000;
  D0 = 0x00000;
  D1 = 0x40000;
  for (; d > 0; d--) {
    d0 = D0;
    d1 = D1;
    for (c = 0x3fff; c >= 0x0000; c--) {
      for (i = 0; i < 16; i++) {
        calc_crc(rom[d0 + i]);
      }
      d0 += 16;
      for (i = 0; i < 16; i++) {
        calc_crc(rom[d1 + i]);
      }
      d1 += 16;
    }
    D0 += 0x80000;
    D1 += 0x80000;
  }
  a = crc;
  a = ((a | 0xf0000) + 1) & 0xfffff;

  if (a == 0x00000) {
    printf("IROM OK: ROM CRC test passed.\n");
  } else {
    printf("IROM %.4x: ROM CRC test FAILED !!!\n", a & 0xffff);
  }

  return 0;
}

