/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: disasm.c,v $
 * Revision 1.1  1994/10/09  20:29:47  ecd
 * Initial revision
 *
 *
 * $Id: disasm.c,v 1.1 1994/10/09 20:29:47 ecd Exp ecd $
 */

#include "global.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "hp48.h"
#include "disasm.h"

int disassembler_mode = HP_MNEMONICS;

static char *opcode_0_tbl[32] = {
  /*
   * HP Mnemonics
   */
  "RTNSXM", "RTN", "RTNSC", "RTNCC",
  "SETHEX", "SETDEC", "RSTK=C", "C=RSTK",
  "CLRST", "C=ST", "ST=C", "CSTEX",
  "P=P+1", "P=P-1", "(NULL)", "RTI",
  /*
   * Alonzo Mnemonics
   */
  "retsetxm", "ret", "retsetc", "retclrc",
  "sethex", "setdec", "push.a    c", "pop.a     c",
  "clr.x     st", "move.x    st, c", "move.x    c, st", "swap.x    c, st",
  "inc.1     p", "dec.1     p", "(null)", "reti"
};

static char *op_str_0[16] = {
  /*
   * HP Mnemonics
   */
  "A=A%cB", "B=B%cC", "C=C%cA", "D=D%cC",
  "B=B%cA", "C=C%cB", "A=A%cC", "C=C%cD",
  /*
   * Alonzo Mnemonics
   */
  "b, a", "c, b", "a, c", "c, d",
  "a, b", "b, c", "c, a", "d, c"
};

static char *field_tbl[32] = {
  /*
   * HP Mnemonics
   */
  "P", "WP", "XS", "X",
  "S", "M", "B", "W",
  "(NULL)", "(NULL)", "(NULL)", "(NULL)", 
  "(NULL)", "(NULL)", "(NULL)", "A",
  /*
   * Alonzo Mnemonics
   */
  ".p", ".wp", ".xs", ".x",
  ".s", ".m", ".b", ".w",
  "(null)", "(null)", "(null)", "(null)", 
  "(null)", "(null)", "(null)", ".a",

};

char *
#ifdef __FunctionProto__
append_str(char *buf, char *str)
#else
append_str(buf, str)
char *buf;
char *str;
#endif
{
  while ((*buf = *str++)) buf++;
  return buf;
}

char *
#ifdef __FunctionProto__
append_tab(char *buf)
#else
append_tab(buf)
char *buf;
#endif
{
  int n;
  char *p;

  n = 10 - (strlen(buf) % 10);
  p = &buf[strlen(buf)];
  while (n--) *p++ = ' ';
  return p;
}

char *
#ifdef __FunctionProto__
append_field(char *buf, word_4 fn)
#else
append_field(buf, fn)
char *buf;
word_4 fn;
#endif
{
  buf = append_str(buf, field_tbl[fn + 16 * disassembler_mode]);
  return buf;
}

word_20
#ifdef __FunctionProto__
Instruction(word_20 addr, char *out)
#else
Instruction(addr, out)
word_20 addr;
char *out;
#endif
{
  word_4  n;
  word_4  fn;
  char   *p = out;
  char    c;
  char    buf[20];
  word_20 disp, pc;

  switch (n = read_nibble(addr++)) {
    case 0:
      if ((n = read_nibble(addr++)) != 0xe) {
        p = append_str(out, opcode_0_tbl[n + 16 * disassembler_mode]);
        break;
      }
      fn = read_nibble(addr++);
      n = read_nibble(addr++);
      switch (disassembler_mode) {
        case HP_MNEMONICS:
          sprintf(buf, op_str_0[(n & 7) + 8 * HP_MNEMONICS],
                  (n < 8) ? '&' : '!');
          p = append_str(out, buf);
          p = append_tab(out);
          p = append_field(p, fn);
          break;
        case ALONZO_MNEMONICS:
          p = append_str(out, (n < 8) ? "and" : "or");
          p = append_field(p, fn);
          p = append_tab(out);
          p = append_str(p, op_str_0[(n & 7) + 8 * ALONZO_MNEMONICS]);
          break;
        default:
          p = append_str(out, "Unknown disassembler mode");
          break;
      }
      break;

    case 2:
      n = read_nibble(addr++);
      switch (disassembler_mode) {
        case HP_MNEMONICS:
          sprintf(buf, "P=%d", n);
          p = append_str(out, buf);
          break;
        case ALONZO_MNEMONICS:
          sprintf(buf, "move.1    %d, p", n);
          p = append_str(out, buf);
          break;
        default:
          p = append_str(out, "Unknown disassembler mode");
          break;
      }
      break;
      
    case 4:
      break;

    default:
      break;
  }
  *p = '\0';
  return addr;
}

