/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: dump2rom.c,v $
 * Revision 1.3  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.2  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.1  1994/09/07  12:57:36  ecd
 * Initial revision
 *
 * $Id: dump2rom.c,v 1.3 1994/09/18 15:29:22 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>

unsigned char *core;

int
#ifdef __FunctionProto__
main(int argc, char **argv)
#else
main(argc, argv)
int argc;
char **argv;
#endif
{
  FILE *dump;
  long a, addr;
  unsigned char ch;
  int i, out;

  if (argc < 2) {
    fprintf(stderr, "usage: %s hp48-dump-file\n", argv[0]);
    exit (1);
  }
  if ((dump = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "%s: can\'t open %s\n", argv[0], argv[1]);
    exit (1);
  }
  if ((core = (unsigned char *)malloc(0x80000)) == NULL) {
    fprintf(stderr, "%s: can\'t malloc %d bytes\n", argv[0], 0x80000);
    exit (1);
  }
  addr = 0;
  while (addr < 0x80000) {
    a = 0;
    for (i = 0; i < 5; i++) {
      a <<= 4;
      ch = fgetc(dump);
      if (ch >= '0' && ch <= '9') {
        a |= ch - '0';
      } else if (ch >= 'A' && ch <= 'F') {
        a |= ch - 'A' + 10;
      } else {
        fprintf(stderr, "%s: Illegal char %c at %lx\n", argv[0], ch, addr);
        exit (1);
      }
    }
    if (a != addr) {
      fprintf(stderr, "%s: Illegal addr %lx, expected %lx\n",
                      argv[0], a, addr);
      exit (1);
    }
    ch = fgetc(dump);
    if (ch != ':') {
      fprintf(stderr, "%s: Illegal char %c, expected \':\' at %lx\n",
                      argv[0], ch, addr);
      exit (1);
    }
    for (i = 0; i < 16; i++) {
      ch = fgetc(dump);
      if (ch >= '0' && ch <= '9') {
        core[addr++] = ch - '0';
      } else if (ch >= 'A' && ch <= 'F') {
        core[addr++] = ch - 'A' + 10;
      } else {
        fprintf(stderr, "%s: Illegal char %c at %lx\n", argv[0], ch, addr);
        exit (1);
      }
    }
    ch = fgetc(dump);
    if (ch != '\n') {
      fprintf(stderr, "%s: Illegal char %c, expected \'\\n\' at %lx\n",
                      argv[0], ch, addr);
      exit (1);
    }
  }
  if ((out = open("rom.dump", O_WRONLY | O_CREAT, 0666)) < 0) {
    fprintf(stderr, "%s: can\'t open %s\n", argv[0], "rom.dump");
    exit (1);
  }
  if (write(out, core, 0x80000) != 0x80000) {
    fprintf(stderr, "%s: can\'t write %d bytes\n", argv[0], 0x80000);
    exit (1);
  }
  exit (0);
}

