/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: main.c,v $
 * Revision 1.6  1994/10/05  08:36:44  ecd
 * new function call to init_nibble_maps()
 *
 * Revision 1.5  1994/09/30  12:37:09  ecd
 * deleted saturn.intenable = 1; statement
 *
 * Revision 1.4  1994/09/18  22:47:20  ecd
 * fixed typo
 *
 * Revision 1.3  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.2  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.1  1994/09/13  15:05:05  ecd
 * Initial revision
 *
 * $Id: main.c,v 1.6 1994/10/05 08:36:44 ecd Exp ecd $
 */


#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <fcntl.h>

#include "x48_x11.h"
#include "hp48.h"
#include "debugger.h"

char *progname;

saturn_t saturn;

extern int init_annunc __ProtoType__((void));
extern int init_nibble_maps __ProtoType__((void));

void
#ifdef __FunctionProto__
signal_handler(int sig)
#else
signal_handler(sig)
int sig;
#endif
{
  switch (sig) {
    case SIGINT:
      enter_debugger |= USER_INTERRUPT;
      break;
#if 0
#if defined(SIGIO) && defined(FASYNC) && defined(F_SETOWN)
    case SIGIO:
      got_alarm = 1;
      break;
#endif
#endif
    case SIGALRM:
      got_alarm = 1;
      break;
    default:
      break;
  }
}

int
#ifdef __FunctionProto__
main(int argc, char **argv)
#else
main(argc, argv)
int argc;
char **argv;
#endif
{
  char *name;
  sigset_t set;
  struct sigaction sa;
#if 0
#if defined(SIGIO) && defined(FASYNC) && defined(F_SETOWN)
  int fd;
  long flags;
#endif
#else
  struct itimerval it;
#endif

  name = (char *)0;
  /*
   *  Get the name we are called.
   */
  progname = strrchr(argv[0], '/');
  if (progname == NULL)
    progname = argv[0];
  else
    progname++;

  /*
   *  Open up the display
   */
  if (OpenDisplay(name) < 0) {
    fprintf(stderr, "%s: can\'t open display\n", progname);
    exit (1);
  }

  /*
   * initialize emulator stuff
   */
  init_emulator();
  
  /*
   *  Create the HP-48 window
   */
  if (CreateMainWindow() < 0) {
    fprintf(stderr, "%s: can\'t create window\n", progname);
    exit (1);
  }

  /*
   * can't be done before windows exist
   */
  init_annunc();
  init_nibble_maps();
  init_display();

  /*
   * Flush the event queue
   */
  GetEvent();

  /*
   *  install a handler for SIGALRM
   */
  sigemptyset(&set);
  sigaddset(&set, SIGALRM);
  sa.sa_handler = signal_handler;
  sa.sa_mask = set;
#ifdef SA_RESTART
  sa.sa_flags = SA_RESTART;
#endif
  sigaction(SIGALRM, &sa, (struct sigaction *)0);

  /*
   *  install a handler for SIGINT
   */
  sigemptyset(&set);
  sigaddset(&set, SIGINT);
  sa.sa_handler = signal_handler;
  sa.sa_mask = set;
#ifdef SA_RESTART
  sa.sa_flags = SA_RESTART;
#endif
  sigaction(SIGINT, &sa, (struct sigaction *)0);

#if 0
#if defined(SIGIO) && defined(FASYNC) && defined(F_SETOWN)

  /*
   *  install a handler for SIGIO
   */
  sigemptyset(&set);
  sigaddset(&set, SIGIO);
  sa.sa_handler = signal_handler;
  sa.sa_mask = set;
#ifdef SA_RESTART
  sa.sa_flags = SA_RESTART;
#endif
  sigaction(SIGIO, &sa, (struct sigaction *)0);
  siginterrupt(SIGIO, 0);

  fd = dpy->fd;
  flags = fcntl(fd, F_GETFL, 0);
  flags |= FASYNC;
  fcntl(fd, F_SETFL, flags);
  fcntl(fd, F_SETOWN, getpid());

#endif
#else

  /*
   * set the real time interval timer
   */
  it.it_interval.tv_sec = 0;
  it.it_interval.tv_usec = 50000;
  it.it_value.tv_sec = 0;
  it.it_value.tv_usec = 50000;
  setitimer(ITIMER_REAL, &it, (struct itimerval *)0);

#endif

  do {

    emulate();

    debug();

  } while (1);

  return 0;
}

