/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: serial.c,v $
 * Revision 1.7  1994/10/06  16:30:05  ecd
 * new init for IRIX
 * added CREAD for serial line
 *
 * Revision 1.6  1994/10/05  08:49:59  ecd
 * changed printf() to print the correct /dev/ttyp?
 *
 * Revision 1.5  1994/09/30  12:37:09  ecd
 * check if serial device is opened by OPENIO
 *
 * Revision 1.4  1994/09/18  15:29:22  ecd
 * turned off unused rcsid message
 *
 * Revision 1.3  1994/09/13  16:57:00  ecd
 * changed to plain X11
 *
 * Revision 1.2  1994/08/31  18:23:21  ecd
 * changed IR and wire definitions.
 *
 * Revision 1.1  1994/08/26  11:09:02  ecd
 * Initial revision
 *
 * $Id: serial.c,v 1.7 1994/10/06 16:30:05 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/time.h>
#ifdef LINUX
#include <sys/ioctl.h>
#endif
#include <unistd.h>
#include <termios.h>
#ifdef SOLARIS
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/termios.h>
#endif

#include "hp48.h"
#include "device.h"
#include "hp48_emu.h"

static int wire_fd;
static int ir_fd;

extern int rece_instr;

/* #define DEBUG_SERIAL */

int
#ifdef __FunctionProto__
serial_init(void)
#else
serial_init()
#endif
{
  int n;
  int ttyp;
#if defined(SOLARIS) || defined (IRIX)
  char *p;
#endif
  char ttyname[128];
  struct termios ttybuf;

  wire_fd = -1;
  ttyp = -1;
#ifdef IRIX
  if ((p = _getpty(&wire_fd, O_RDWR | O_EXCL | O_NDELAY, 0666, 0)) == NULL) {
    wire_fd = -1;
  } else {
    if ((ttyp = open(p, O_RDWR | O_NDELAY, 0666)) < 0) {
      close(wire_fd);
      wire_fd = -1;
    } else {
      fprintf(stdout, "wire connection on %s\n", p);
    }
  }
#else  /* IRIX */
#ifdef SOLARIS
  if ((wire_fd = open("/dev/ptmx", O_RDWR | O_NONBLOCK, 0666)) >= 0) {
    grantpt(wire_fd);
    unlockpt(wire_fd);
    p = ptsname(wire_fd);
    strcpy(ttyname, p);
    if ((ttyp = open(ttyname, O_RDWR | O_NDELAY, 0666)) >= 0) {
      ioctl(ttyp, I_PUSH, "ptem");
      ioctl(ttyp, I_PUSH, "ldterm");
      fprintf(stdout, "wire connection on %s\n", ttyname);
    }
  }
#else  /* SOLARIS */
#ifdef LINUX
  for (n = 0; n < 16; n++) {
    sprintf(ttyname, "/dev/ptyp%x", n);
    if ((wire_fd = open(ttyname, O_RDWR | O_EXCL | O_NDELAY, 0666)) >= 0) {
      ttyp = wire_fd;
      sprintf(ttyname, "/dev/ttyp%x", n);
      fprintf(stdout, "wire connection on %s\n", ttyname);
      break;
    }
  }
#else  /* LINUX */
  for (n = 0; n < 16; n++) {
    sprintf(ttyname, "/dev/ptyp%x", n);
    if ((wire_fd = open(ttyname, O_RDWR | O_EXCL | O_NDELAY, 0666)) >= 0) {
      sprintf(ttyname, "/dev/ttyp%x", n);
      if ((ttyp = open(ttyname, O_RDWR | O_NDELAY, 0666)) < 0) {
        wire_fd = -1;
      }
      fprintf(stdout, "wire connection on %s\n", ttyname);
      break;
    }
  }
#endif /* LINUX */
#endif /* SOLARIS */
#endif /* IRIX */
  if (ttyp >= 0) {
#ifdef SUNOS
    if (tcgetattr(ttyp, &ttybuf) < 0) {
#else
    if (ioctl(ttyp, TCGETS, (char *)&ttybuf) < 0) {
#endif
      fprintf(stderr, "ioctl TCGETS failed, errno = %d\n", errno);
      wire_fd = -1;
    }
    ttybuf.c_lflag = 0;
    ttybuf.c_iflag = 0;
    ttybuf.c_oflag = 0;
    ttybuf.c_cflag = B9600 | CS8 | CLOCAL;
    for (n = 0; n < NCCS; n++)
      ttybuf.c_cc[n] = 0;
    ttybuf.c_cc[VTIME] = 0;
    ttybuf.c_cc[VMIN] = 1;
#ifdef SUNOS
    if (tcsetattr(ttyp, TCSANOW, &ttybuf) < 0) {
#else
    if (ioctl(ttyp, TCSETS, (char *)&ttybuf) < 0) {
#endif
      fprintf(stderr, "ioctl TCSETS failed, errno = %d\n", errno);
      wire_fd = -1;
    }
  }
#ifdef IRIX
  sprintf(ttyname, "/dev/ttyd2");
#else  /* IRIS */
#ifdef SOLARIS
  sprintf(ttyname, "/dev/ttya");
#else  /* SOLARIS */
#ifdef LINUX
  sprintf(ttyname, "/dev/ttyS7");
#else  /* LINUX */
  sprintf(ttyname, "/dev/ttya");
#endif /* LINUX */
#endif /* SOLARIS */
#endif /* IRIX */
  ir_fd = -1;
  if ((ir_fd = open(ttyname, O_RDWR | O_NDELAY)) >= 0) {
    fprintf(stdout, "IR connection on %s\n", ttyname);
#ifdef SUNOS
    if (tcgetattr(ir_fd, &ttybuf) < 0) {
#else
    if (ioctl(ir_fd, TCGETS, (char *)&ttybuf) < 0) {
#endif
      fprintf(stderr, "ioctl TCGETS failed, errno = %d\n", errno);
      ir_fd = -1;
    }
    ttybuf.c_lflag = 0;
    ttybuf.c_iflag = 0;
    ttybuf.c_oflag = 0;
    ttybuf.c_cflag = B9600 | CS8 | CLOCAL | CREAD;
    for (n = 0; n < NCCS; n++)
      ttybuf.c_cc[n] = 0;
    ttybuf.c_cc[VTIME] = 0;
    ttybuf.c_cc[VMIN] = 1;
#ifdef SUNOS
    if (tcsetattr(ir_fd, TCSANOW, &ttybuf) < 0) {
#else
    if (ioctl(ir_fd, TCSETS, (char *)&ttybuf) < 0) {
#endif
      fprintf(stderr, "ioctl TCSETS failed, errno = %d\n", errno);
      ir_fd = -1;
    }
  }
  return 1;
}

void
#ifdef __FunctionProto__
transmit_char(void)
#else
transmit_char()
#endif
{
#ifdef DEBUG_SERIALx
  fprintf(stderr, "XMT %s\n", (saturn.ir_ctrl & 0x04) ? "IR" : "wire");
#endif

  if (saturn.ir_ctrl & 0x04) {
    if (ir_fd == -1) {
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
      return;
    }
  } else {
    if (wire_fd == -1) {
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
      return;
    }
  }

#ifdef DEBUG_SERIAL
  if (isprint(saturn.tbr)) {
    fprintf(stderr, "-> \'%c\'\n", saturn.tbr);
  } else {
    fprintf(stderr, "-> %x\n", saturn.tbr);
  }
#endif
  if (saturn.ir_ctrl & 0x04) {
    if (write(ir_fd, &saturn.tbr, 1) == 1) {
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
    } else {
      if (errno != EAGAIN) {
        fprintf(stderr, "Serial write error: %d\n", errno);
      }
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
    }
  } else {
    if (write(wire_fd, &saturn.tbr, 1) == 1) {
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
    } else {
      if (errno != EAGAIN) {
        fprintf(stderr, "Serial write error: %d\n", errno);
      }
      saturn.tcs &= 0x0e;
      if (saturn.io_ctrl & 0x04) {
        do_interupt();
      }
    }
  }
}

#define NR_BUFFER 256

void
#ifdef __FunctionProto__
receive_char()
#else
receive_char()
#endif
{
  struct timeval tout;
  fd_set rfds;
  int nfd;
  static unsigned char buf[NR_BUFFER + 1];
  static int nrd = 0, bp = 0;

#ifdef DEBUG_SERIALx
  fprintf(stderr, "RCV %s\n", (saturn.ir_ctrl & 0x04) ? "IR" : "wire");
#endif

  rece_instr = 0;

  if (saturn.ir_ctrl & 0x04) {
    if (ir_fd == -1)
      return;
  } else {
    if (wire_fd == -1)
      return;
  }

  if (saturn.rcs & 0x01) {
    return;
  }

  if (nrd == 0) {
    tout.tv_sec = 0;
    tout.tv_usec = 0;
    FD_ZERO(&rfds);
    if (saturn.ir_ctrl & 0x04) {
      FD_SET(ir_fd, &rfds);
      nfd = ir_fd + 1;
    } else {
      FD_SET(wire_fd, &rfds);
      nfd = wire_fd + 1;
    }
    if ((nfd = select(nfd, &rfds, (fd_set *)0, (fd_set *)0, &tout)) > 0) {
#ifdef DEBUG_SERIAL
      fprintf(stderr, "select = %d\n", nfd);
#endif
      if (saturn.ir_ctrl & 0x04) {
        if (FD_ISSET(ir_fd, &rfds)) {
          nrd = read(ir_fd, buf, NR_BUFFER);
          if (nrd < 0) {
            nrd = 0;
            return;
          }
          bp = 0;
        } else {
          return;
        }
      } else {
        if (FD_ISSET(wire_fd, &rfds)) {
          nrd = read(wire_fd, buf, NR_BUFFER);
          if (nrd < 0) {
            nrd = 0;
            return;
          }
          bp = 0;
        } else {
          return;
        }
      }
    } else {
      return;
    }
  }
  if (nrd == 0) {
    return;
  }
  if (!(saturn.io_ctrl & 0x08)) {
    nrd = 0;
    return;
  }
  saturn.rbr = buf[bp++];
  nrd--;
  saturn.rcs |= 0x01;
  if (saturn.io_ctrl & 0x02) {
    do_interupt();
  }
}

