
#include "global.h"

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/time.h>

#include "timer.h"

/* #define DEBUG_TIMER 1 */

#ifdef SOLARIS
extern int gettimeofday __ProtoType__((struct timeval *tp));
#endif
#ifdef SUNOS
extern int gettimeofday __ProtoType__((struct timeval *, struct timezone *));
#endif

typedef struct x48_timer_t {
  word_1  run;
  word_64 start;
  word_64 stop;
  word_64 value;
} x48_timer_t;

static x48_timer_t timers[NR_TIMERS];

long
#ifdef __FunctionProto__
diff_timer(word_64 *t1, word_64 *t2)
#else
diff_timer(t1, t2)
word_64 *t1;
word_64 *t2;
#endif
{
  return (t1->lo - t2->lo);
}

static inline void
#ifdef __FunctionProto__
add_sub_64(word_64 *t1, word_64 *t2, word_64 *diff)
#else
add_sub_64(t1, t2, diff)
word_64 *t1;
word_64 *t2;
word_64 *diff;
#endif
{
  unsigned short s1[4], s2[4];
  unsigned long  r[4];

  s1[0] = (unsigned short)(t1->lo & 0xffff);
  s1[1] = (unsigned short)((t1->lo >> 16) & 0xffff);
  s1[2] = (unsigned short)(t1->hi & 0xffff);
  s1[3] = (unsigned short)((t1->hi >> 16) & 0xffff);

  s2[0] = (unsigned short)(t2->lo & 0xffff);
  s2[1] = (unsigned short)((t2->lo >> 16) & 0xffff);
  s2[2] = (unsigned short)(t2->hi & 0xffff);
  s2[3] = (unsigned short)((t2->hi >> 16) & 0xffff);

  r[0] = (unsigned long)s1[0] - (unsigned long)s2[0];
  r[1] = (unsigned long)s1[1] - (unsigned long)s2[1];
  r[2] = (unsigned long)s1[2] - (unsigned long)s2[2];
  r[3] = (unsigned long)s1[3] - (unsigned long)s2[3];

  if (r[0] >> 16) r[1]--;
  if (r[1] >> 16) r[2]--;
  if (r[2] >> 16) r[3]--;

  s1[0] = (unsigned short)(diff->lo & 0xffff);
  s1[1] = (unsigned short)((diff->lo >> 16) & 0xffff);
  s1[2] = (unsigned short)(diff->hi & 0xffff);
  s1[3] = (unsigned short)((diff->hi >> 16) & 0xffff);

  s2[0] = (unsigned short)r[0];
  s2[1] = (unsigned short)r[1];
  s2[2] = (unsigned short)r[2];
  s2[3] = (unsigned short)r[3];

  r[0] = (unsigned long)s1[0] + (unsigned long)s2[0];
  r[1] = (unsigned long)s1[1] + (unsigned long)s2[1];
  r[2] = (unsigned long)s1[2] + (unsigned long)s2[2];
  r[3] = (unsigned long)s1[3] + (unsigned long)s2[3];

  if (r[0] >> 16) r[1]++;
  if (r[1] >> 16) r[2]++;
  if (r[2] >> 16) r[3]++;

  t2->lo = t1->lo;
  t2->hi = t1->hi;

  diff->lo = ((r[1] << 16) | (r[0] & 0xffff));
  diff->hi = ((r[3] << 16) | (r[2] & 0xffff));
}

void
#ifdef __FunctionProto__
start_timer(int timer)
#else
start_timer(timer)
int timer;
#endif
{
  struct timeval tv;

  if (timer > NR_TIMERS)
    return;

  if (timers[timer].run == 1)
    return;

#ifdef SOLARIS
  gettimeofday(&tv);
#else
  gettimeofday(&tv, (struct timezone *)0);
#endif

  timers[timer].run = 1;
  if (timer == T1_TIMER) {
    timers[timer].start.hi = (tv.tv_sec >> 23);
    timers[timer].start.lo = (tv.tv_sec << 9);
    timers[timer].start.lo |= tv.tv_usec / 62500;
  } else {
    timers[timer].start.hi = (tv.tv_sec >> 19);
    timers[timer].start.lo = (tv.tv_sec << 13);
    timers[timer].start.lo |= (tv.tv_usec << 7) / 15625;
  }
#ifdef DEBUG_TIMER
  fprintf(stderr, "Timer[%d] start at 0x%.8lx%.8lx\n", timer,
          timers[timer].start.hi, timers[timer].start.lo);
#endif
}

void
#ifdef __FunctionProto__
restart_timer(int timer)
#else
restart_timer(timer)
int timer;
#endif
{
  struct timeval tv;

  if (timer > NR_TIMERS)
    return;

  timers[timer].start.lo = timers[timer].start.hi = 0;
  timers[timer].stop.lo = timers[timer].stop.hi = 0;
  timers[timer].value.lo = timers[timer].value.hi = 0;

#ifdef SOLARIS
  gettimeofday(&tv);
#else
  gettimeofday(&tv, (struct timezone *)0);
#endif

  timers[timer].run = 1;
  if (timer == T1_TIMER) {
    timers[timer].start.hi = (tv.tv_sec >> 23);
    timers[timer].start.lo = (tv.tv_sec << 9);
    timers[timer].start.lo |= tv.tv_usec / 62500;
  } else {
    timers[timer].start.hi = (tv.tv_sec >> 19);
    timers[timer].start.lo = (tv.tv_sec << 13);
    timers[timer].start.lo |= (tv.tv_usec << 7) / 15625;
  }
#ifdef DEBUG_TIMER
  fprintf(stderr, "Timer[%d] restart at 0x%.8lx%.8lx\n", timer,
          timers[timer].start.hi, timers[timer].start.lo);
#endif
}

void
#ifdef __FunctionProto__
stop_timer(int timer)
#else
stop_timer(timer)
int timer;
#endif
{
  struct timeval tv;

  if (timer > NR_TIMERS)
    return;

  if (timers[timer].run == 0)
    return;

#ifdef SOLARIS
  gettimeofday(&tv);
#else
  gettimeofday(&tv, (struct timezone *)0);
#endif

  timers[timer].run = 0;
  if (timer == T1_TIMER) {
    timers[timer].stop.hi = (tv.tv_sec >> 23);
    timers[timer].stop.lo = (tv.tv_sec << 9);
    timers[timer].stop.lo |= tv.tv_usec / 62500;
  } else {
    timers[timer].stop.hi = (tv.tv_sec >> 19);
    timers[timer].stop.lo = (tv.tv_sec << 13);
    timers[timer].stop.lo |= (tv.tv_usec << 7) / 15625;
  }
  add_sub_64(&timers[timer].stop, &timers[timer].start, &timers[timer].value);
#ifdef DEBUG_TIMER
  fprintf(stderr, "Timer[%d] stop at 0x%.8lx%.8lx, value 0x%.8lx%.8lx\n",
          timer, timers[timer].stop.hi, timers[timer].stop.lo,
          timers[timer].value.hi, timers[timer].value.lo);
#endif
}

void
#ifdef __FunctionProto__
reset_timer(int timer)
#else
reset_timer(timer)
int timer;
#endif
{
  if (timer > NR_TIMERS)
    return;

  timers[timer].run = 0;
  timers[timer].start.lo = timers[timer].start.hi = 0;
  timers[timer].stop.lo = timers[timer].stop.hi = 0;
  timers[timer].value.lo = timers[timer].value.hi = 0;
#ifdef DEBUG_TIMER
  fprintf(stderr, "Timer[%d] reset\n", timer);
#endif
}

static word_64 zero = { 0, 0 };

word_64
#ifdef __FunctionProto__
get_timer(int timer)
#else
get_timer(timer)
int timer;
#endif
{
  struct timeval tv;
  word_64 stop;

  if (timer > NR_TIMERS)
    return zero;

  if (timers[timer].run) {

#ifdef SOLARIS
    gettimeofday(&tv);
#else
    gettimeofday(&tv, (struct timezone *)0);
#endif

    if (timer == T1_TIMER) {
      stop.hi = (tv.tv_sec >> 23);
      stop.lo = (tv.tv_sec << 9);
      stop.lo |= tv.tv_usec / 62500;
    } else {
      stop.hi = (tv.tv_sec >> 19);
      stop.lo = (tv.tv_sec << 13);
      stop.lo |= (tv.tv_usec << 7) / 15625;
    }
    add_sub_64(&stop, &timers[timer].start, &timers[timer].value);
  }
  return timers[timer].value;
}


t1_t2_ticks
#ifdef __FunctionProto__
get_t1_t2(void)
#else
get_t1_t2()
#endif
{
  struct timeval tv;
  word_64 stop;
  t1_t2_ticks ticks;

  if (timers[T1_TIMER].run || timers[T2_TIMER].run) {
#ifdef SOLARIS
    gettimeofday(&tv);
#else
    gettimeofday(&tv, (struct timezone *)0);
#endif
  }
  if (timers[T1_TIMER].run) {
    stop.hi = (tv.tv_sec >> 23);
    stop.lo = (tv.tv_sec << 9);
    stop.lo |= tv.tv_usec / 62500;
    add_sub_64(&stop, &timers[T1_TIMER].start, &timers[T1_TIMER].value);
  }
  if (timers[T2_TIMER].run) {
    stop.hi = (tv.tv_sec >> 19);
    stop.lo = (tv.tv_sec << 13);
    stop.lo |= (tv.tv_usec << 7) / 15625;
    add_sub_64(&stop, &timers[T2_TIMER].start, &timers[T2_TIMER].value);
  }
  ticks.t1_ticks = timers[T1_TIMER].value.lo;
  ticks.t2_ticks = timers[T2_TIMER].value.lo;
  return ticks;
}

