/*
 *  This file is part of x48, an emulator of the HP-48sx Calculator.
 *  Copyright (C) 1994  Eddie C. Dost  (ecd@dressler.de)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/* $Log: x48_x11.c,v $
 * Revision 1.7  1994/10/09  20:32:02  ecd
 * changed refresh_display to support bit offset.
 *
 * Revision 1.6  1994/10/06  16:30:05  ecd
 * added XShm - Extension stuff
 *
 * Revision 1.5  1994/10/05  08:36:44  ecd
 * added backing_store = Always for subwindows
 *
 * Revision 1.4  1994/09/30  12:37:09  ecd
 * added support for interrupt detection in GetEvent,
 * faster display updates,
 * update display window only when mapped.
 *
 * Revision 1.3  1994/09/18  22:47:20  ecd
 * added version information
 *
 * Revision 1.2  1994/09/18  15:29:22  ecd
 * started Real Time support
 *
 * Revision 1.1  1994/09/13  15:05:05  ecd
 * Initial revision
 *
 * $Id: x48_x11.c,v 1.7 1994/10/09 20:32:02 ecd Exp ecd $
 */


#include "global.h"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#ifdef SUNOS
#include <memory.h>
#endif

#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xutil.h>

#include "x48_x11.h"
#include "icon.h"
#include "small.h"
#include "buttons.h"
#include "hp.h"

#include "hp48.h"
#include "device.h"

extern int XShmQueryExtension __ProtoType__((Display *));

Display *dpy;
int	 screen;
int	 depth;
Colormap cmap;
GC	 gc;
Window   mainW;
disp_t   disp;
Atom	 wm_delete_window, wm_protocols;

#ifdef USE_SHM
int      shm_flag;
int	 xerror_flag;
#endif

/* #define DEBUG_XEVENT 1 */
#define DEBUG_SHM 1
#define DEBUG_XIMAGE 1

extern char *progname;

typedef struct keypad_t {
  int width;
  int height;
  Pixmap pixmap;
} keypad_t;

keypad_t keypad;

color_t colors[] = {
  { "white", 255, 255, 255, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "left", 255, 166, 0, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "right", 0, 210, 255, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "but_top", 109, 93, 93, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "button", 90, 77, 77, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "but_bot", 76, 65, 65, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "lcd_col", 202, 221, 92, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "pix_col", 0, 0, 128, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "pad_top", 109, 78, 78, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "pad", 90, 64, 64, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "pad_bot", 76, 54, 54, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "disp_pad_top", 155, 118, 84, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "disp_pad", 124, 94, 67, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "disp_pad_bot", 100, 75, 53, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "logo", 204, 169, 107, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "logo_back", 64, 64, 64, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "label", 202, 184, 144, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { "black", 0, 0, 0, { 0, 0, 0, 0, DoRed | DoGreen | DoBlue, 0 } },
  { 0 }
};

typedef struct button_t {
  char  *name;

  short  pressed;

  int    code;
  int    x, y;
  int    w, h;

  int	 lc;
  char  *label;
  short  font_size;
  int    lw, lh;
  unsigned char *lb;

  char  *letter;

  char  *left;
  short  is_menu;

  char  *right;
  char  *sub;
  Pixmap map;
  Pixmap down;
  Window xwin;
} button_t;

#define BUTTON_A	0
#define BUTTON_B	1
#define BUTTON_C	2
#define BUTTON_D	3
#define BUTTON_E	4
#define BUTTON_F	5

#define BUTTON_MTH	6
#define BUTTON_PRG	7
#define BUTTON_CST	8
#define BUTTON_VAR	9
#define BUTTON_UP	10
#define BUTTON_NXT	11

#define BUTTON_COLON	12
#define BUTTON_STO	13
#define BUTTON_EVAL	14
#define BUTTON_LEFT	15
#define BUTTON_DOWN	16
#define BUTTON_RIGHT	17

#define BUTTON_SIN	18
#define BUTTON_COS	19
#define BUTTON_TAN	20
#define BUTTON_SQRT	21
#define BUTTON_POWER	22
#define BUTTON_INV	23

#define BUTTON_ENTER	24
#define BUTTON_NEG	25
#define BUTTON_EEX	26
#define BUTTON_DEL	27
#define BUTTON_BS	28

#define BUTTON_ALPHA	29
#define BUTTON_7	30
#define BUTTON_8	31
#define BUTTON_9	32
#define BUTTON_DIV	33

#define BUTTON_SHL	34
#define BUTTON_4	35
#define BUTTON_5	36
#define BUTTON_6	37
#define BUTTON_MUL	38

#define BUTTON_SHR	39
#define BUTTON_1	40
#define BUTTON_2	41
#define BUTTON_3	42
#define BUTTON_MINUS	43

#define BUTTON_ON	44
#define BUTTON_0	45
#define BUTTON_PERIOD	46
#define BUTTON_SPC	47
#define BUTTON_PLUS	48

#define LAST_BUTTON	48

button_t buttons[] = {
  { "A", 0, 0x14,  0, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "A", 0, 0, 0, 0, 0 },
  { "B", 0, 0x84,  50, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "B", 0, 0, 0, 0, 0 },
  { "C", 0, 0x83, 100, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "C", 0, 0, 0, 0, 0 },
  { "D", 0, 0x82, 150, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "D", 0, 0, 0, 0, 0 },
  { "E", 0, 0x81, 200, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "E", 0, 0, 0, 0, 0 },
  { "F", 0, 0x80, 250, 0, 36, 23, WHITE,
     0, 0, menu_label_width, menu_label_height, menu_label_bits,
    "F", 0, 0, 0, 0, 0 },

  { "MTH", 0, 0x24,  0, 50, 36, 26, WHITE, "MTH", 0, 0, 0, 0,
    "G", "PRINT", 1, 0, 0, 0 },
  { "PRG", 0, 0x74,  50, 50, 36, 26, WHITE, "PRG", 0, 0, 0, 0,
    "H", "I/O", 1, 0, 0, 0 },
  { "CST", 0, 0x73, 100, 50, 36, 26, WHITE, "CST", 0, 0, 0, 0,
    "I", "MODES", 1, 0, 0, 0 },
  { "VAR", 0, 0x72, 150, 50, 36, 26, WHITE, "VAR", 0, 0, 0, 0,
    "J", "MEMORY", 1, 0, 0, 0 },
  { "UP", 0, 0x71, 200, 50, 36, 26, WHITE,
     0, 0, up_width, up_height, up_bits,
    "K", "LIBRARY", 1, 0, 0, 0 },
  { "NXT", 0, 0x70, 250, 50, 36, 26, WHITE, "NXT", 0, 0, 0, 0,
    "L", "PREV", 0, 0, 0, 0 },

  { "COLON", 0, 0x04,  0, 100, 36, 26, WHITE,
     0, 0, colon_width, colon_height, colon_bits,
    "M", "UP", 0, "HOME", 0, 0 },
  { "STO", 0, 0x64,  50, 100, 36, 26, WHITE, "STO", 0, 0, 0, 0,
    "N", "DEF", 0, "RCL", 0, 0 },
  { "EVAL", 0, 0x63, 100, 100, 36, 26, WHITE, "EVAL", 0, 0, 0, 0,
    "O", "aQ", 0, "aNUM", 0, 0 },
  { "LEFT", 0, 0x62, 150, 100, 36, 26, WHITE,
     0, 0, left_width, left_height, left_bits,
    "P", "GRAPH", 0, 0, 0, 0 },
  { "DOWN", 0, 0x61, 200, 100, 36, 26, WHITE,
     0, 0, down_width, down_height, down_bits,
    "Q", "REVIEW", 0, 0, 0, 0 },
  { "RIGHT", 0, 0x60, 250, 100, 36, 26, WHITE,
     0, 0, right_width, right_height, right_bits,
    "R", "SWAP", 0, 0, 0, 0 },

  { "SIN", 0, 0x34,  0, 150, 36, 26, WHITE, "SIN", 0, 0, 0, 0,
    "S", "ASIN", 0, "b", 0, 0 },
  { "COS", 0, 0x54,  50, 150, 36, 26, WHITE, "COS", 0, 0, 0, 0,
    "T", "ACOS", 0, "c", 0, 0 },
  { "TAN", 0, 0x53,  100, 150, 36, 26, WHITE, "TAN", 0, 0, 0, 0,
    "U", "ATAN", 0, "d", 0, 0 },
  { "SQRT", 0, 0x52,  150, 150, 36, 26, WHITE,
     0, 0, sqrt_width, sqrt_height, sqrt_bits,
    "V", "e", 0, "f", 0, 0 },
  { "POWER", 0, 0x51,  200, 150, 36, 26, WHITE,
     0, 0, power_width, power_height, power_bits,
    "W", "g", 0, "LOG", 0, 0 },
  { "INV", 0, 0x50,  250, 150, 36, 26, WHITE,
     0, 0, inv_width, inv_height, inv_bits,
    "X", "h", 0, "LN", 0, 0 },

  { "ENTER", 0, 0x44, 0, 200, 86, 26, WHITE, "ENTER", 2, 0, 0, 0,
     0, "EQUATION", 0, "MATRIX", 0, 0 },
  { "NEG", 0, 0x43, 100, 200, 36, 26, WHITE,
     0, 0, neg_width, neg_height, neg_bits,
    "Y", "EDIT", 0, "VISIT", 0, 0 },
  { "EEX", 0, 0x42, 150, 200, 36, 26, WHITE, "EEX", 0, 0, 0, 0,
    "Z", "2D", 0, "3D", 0, 0 },
  { "DEL", 0, 0x41, 200, 200, 36, 26, WHITE, "DEL", 0, 0, 0, 0,
     0, "PURGE", 0, 0, 0, 0 },
  { "BS", 0, 0x40, 250, 200, 36, 26, WHITE,
     0, 0, bs_width, bs_height, bs_bits,
     0, "DROP", 0, "CLR", 0, 0 },

  { "ALPHA", 0, 0x35, 0, 250, 36, 26, WHITE,
     0, 0, alpha_width, alpha_height, alpha_bits,
     0, "USR", 0, "ENTRY", 0, 0 },
  { "7", 0, 0x33, 60, 250, 46, 26, WHITE, "7", 1, 0, 0, 0,
     0, "SOLVE", 1, 0, 0, 0 },
  { "8", 0, 0x32, 120, 250, 46, 26, WHITE, "8", 1, 0, 0, 0,
     0, "PLOT", 1, 0, 0, 0 },
  { "9", 0, 0x31, 180, 250, 46, 26, WHITE, "9", 1, 0, 0, 0,
     0, "ALGEBRA", 1, 0, 0, 0 },
  { "DIV", 0, 0x30, 240, 250, 46, 26, WHITE,
     0, 0, div_width, div_height, div_bits,
     0, "( )", 0, "#", 0, 0 },

  { "SHL", 0, 0x25, 0, 300, 36, 26, LEFT,
     0, 0, shl_width, shl_height, shl_bits,
     0, 0, 0, 0, 0, 0 },
  { "4", 0, 0x23, 60, 300, 46, 26, WHITE, "4", 1, 0, 0, 0,
     0, "TIME", 1, 0, 0, 0 },
  { "5", 0, 0x22, 120, 300, 46, 26, WHITE, "5", 1, 0, 0, 0,
     0, "STAT", 1, 0, 0, 0 },
  { "6", 0, 0x21, 180, 300, 46, 26, WHITE, "6", 1, 0, 0, 0,
     0, "UNITS", 1, 0, 0, 0 },
  { "MUL", 0, 0x20, 240, 300, 46, 26, WHITE,
     0, 0, mul_width, mul_height, mul_bits,
     0, "[ ]", 0, "_", 0, 0 },

  { "SHR", 0, 0x15, 0, 350, 36, 26, RIGHT,
     0, 0, shr_width, shr_height, shr_bits,
     0, 0, 0, 0, 0, 0 },
  { "1", 0, 0x13, 60, 350, 46, 26, WHITE, "1", 1, 0, 0, 0,
     0, "RAD", 0, "POLAR", 0, 0 },
  { "2", 0, 0x12, 120, 350, 46, 26, WHITE, "2", 1, 0, 0, 0,
     0, "STACK", 0, "ARG", 0, 0 },
  { "3", 0, 0x11, 180, 350, 46, 26, WHITE, "3", 1, 0, 0, 0,
     0, "CMD", 0, "MENU", 0, 0 },
  { "MINUS", 0, 0x10, 240, 350, 46, 26, WHITE,
     0, 0, minus_width, minus_height, minus_bits,
     0, "i", 0, "j", 0, 0 },

  { "ON", 0, 0x8000, 0, 400, 36, 26, WHITE, "ON", 0, 0, 0, 0,
     0, "CONT", 0, "OFF", "ATTN", 0 },
  { "0", 0, 0x03, 60, 400, 46, 26, WHITE, "0", 1, 0, 0, 0,
     0, "= ", 0, " a", 0, 0 },
  { "PERIOD", 0, 0x02, 120, 400, 46, 26, WHITE, ".", 1, 0, 0, 0,
     0, ", ", 0, " k", 0, 0 },
  { "SPC", 0, 0x01, 180, 400, 46, 26, WHITE, "SPC", 0, 0, 0, 0,
     0, "l ", 0, " m", 0, 0 },
  { "PLUS", 0, 0x00, 240, 400, 46, 26, WHITE,
     0, 0, plus_width, plus_height, plus_bits,
     0, "{ }", 0, ": :", 0, 0 },

  { 0 }
};

#define KEYBOARD_HEIGHT	(buttons[LAST_BUTTON].y + buttons[LAST_BUTTON].h)
#define KEYBOARD_WIDTH  (buttons[LAST_BUTTON].x + buttons[LAST_BUTTON].w)

#define TOP_SKIP		65
#define SIDE_SKIP		20
#define BOTTOM_SKIP		25
#define DISP_KBD_SKIP		65

#define DISPLAY_WIDTH		(264 + 8)
#define DISPLAY_HEIGHT		(128 + 16 + 8)
#define DISPLAY_OFFSET_X	(SIDE_SKIP+(KEYBOARD_WIDTH-DISPLAY_WIDTH)/2)
#define DISPLAY_OFFSET_Y	TOP_SKIP

#define DISP_FRAME		8

#define KEYBOARD_OFFSET_X	SIDE_SKIP
#define KEYBOARD_OFFSET_Y	(TOP_SKIP + DISPLAY_HEIGHT + DISP_KBD_SKIP)

int
#ifdef __FunctionProto__
AllocColors(void)
#else
AllocColors()
#endif
{
  int c, error;

  error = 0;
  for (c = WHITE; c <= BLACK; c++) {
    colors[c].xcolor.red = colors[c].r << 8;
    colors[c].xcolor.green = colors[c].g << 8;
    colors[c].xcolor.blue = colors[c].b << 8;
    if (XAllocColor(dpy, cmap, &colors[c].xcolor) != 0)
      error = 1;
  }
  if (error)
    return -1;
  return 0;
}

int
#ifdef __FunctionProto__
OpenDisplay(char *name)
#else
OpenDisplay(name)
char *name;
#endif
{

  /*
   * Open the display
   */
  dpy = XOpenDisplay(name);
  if (dpy == (Display *)0)
    return -1;

  /*
   * Get the default screen
   */
  screen = DefaultScreen(dpy);

#ifdef USE_SHM
  /*
   * Try to use XShm-Extension
   */
  shm_flag = 1;

  if (!XShmQueryExtension(dpy)) {
    shm_flag = 0;
#ifdef DEBUG_SHM
    fprintf(stderr, "Shared Memory not supported, using normal Xlib.\n");
#endif
  }
#endif

  return 0;
}

int
#ifdef __FunctionProto__
SmallTextWidth(XFontStruct *font_struct, const char *string, int length)
#else
SmallTextWidth(font_struct, string, length)
XFontStruct *font_struct;
const char *string;
int length;
#endif
{
  int i, w;

  w = 0;
  for (i = 0; i < length; i++) {
    if (small_font[(int)string[i]].h != 0) {
      w += small_font[(int)string[i]].w + 1;
    } else {
      w += XTextWidth(font_struct, &string[i], 1);
    }
  }
  return w;
}

int
#ifdef __FunctionProto__
DrawSmallString(Display *dpy, Drawable d, GC gc, XFontStruct *font,
                int x, int y, const char *string, int length)
#else
DrawSmallString(dpy, d, gc, font, x, y, string, length)
Display *dpy;
Drawable d;
GC gc;
XFontStruct *font;
int x;
int y;
const char *string;
int length;
#endif
{
  int i;
  Pixmap pix;

  for (i = 0; i < length; i++) {
    if (small_font[(int)string[i]].h != 0) {
      pix = XCreateBitmapFromData(dpy, d, small_font[(int)string[i]].bits,
		small_font[(int)string[i]].w, small_font[(int)string[i]].h);
      XCopyPlane(dpy, pix, d, gc, 0, 0, small_font[(int)string[i]].w,
                 small_font[(int)string[i]].h, x,
	         y - small_font[(int)string[i]].h, 1);
      XFreePixmap(dpy, pix);
    } else {
      XDrawImageString(dpy, d, gc, x, y, &string[i], 1);
    }
    x += SmallTextWidth(font, &string[i], 1);
  }
  return 0;
}

int
#ifdef __FunctionProto__
CreateButtons(void)
#else
CreateButtons()
#endif
{
  int i, x, y;
  XSetWindowAttributes xswa;
  XFontStruct *finfo, *f_small, *f_med, *f_big;
  XGCValues val;
  unsigned long gc_mask;
  Pixmap pix;
  XCharStruct xchar;
  int dir, fa, fd;
  unsigned long pixel;

  f_small = XLoadQueryFont(dpy, "7x14bold");
  f_med = XLoadQueryFont(dpy, "9x15bold");
  f_big = XLoadQueryFont(dpy, "10x20");

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {

    if (i < BUTTON_MTH)
      pixel = COLOR(DISP_PAD);
    else
      pixel = COLOR(PAD);

    /*
     * create the buttons subwindows
     */
    buttons[i].xwin = XCreateSimpleWindow(dpy, mainW,
			KEYBOARD_OFFSET_X + buttons[i].x,
			KEYBOARD_OFFSET_Y + buttons[i].y,
			buttons[i].w, buttons[i].h, 0,
			BlackPixel(dpy, screen), pixel);

    xswa.event_mask = ExposureMask | StructureNotifyMask;
    xswa.backing_store = Always;

    XChangeWindowAttributes(dpy, buttons[i].xwin,
                            CWEventMask|CWBackingStore, &xswa);

    /*
     * draw the released button
     */
    buttons[i].map = XCreatePixmap(dpy, buttons[i].xwin, buttons[i].w,
				buttons[i].h, depth);

    XSetForeground(dpy, gc, pixel);
    XFillRectangle(dpy, buttons[i].map, gc, 0, 0, buttons[i].w,
                   buttons[i].h);

    XSetForeground(dpy, gc, COLOR(BUTTON));
    XFillRectangle(dpy, buttons[i].map, gc, 1, 1, buttons[i].w - 2,
                   buttons[i].h - 2);

    if (buttons[i].label != (char *)0) {

      /*
       * set font size in gc
       */
      switch (buttons[i].font_size) {
        case 0:
          finfo = f_small;
          break;
        case 1:
          finfo = f_big;
          break;
        case 2:
          finfo = f_med;
          break;
        default:
          finfo = f_small;
          break;
      }
      val.font = finfo->fid;
      gc_mask = GCFont;
      XChangeGC(dpy, gc, gc_mask, &val);

      /*
       * draw string centered in button
       */
      XSetBackground(dpy, gc, COLOR(BUTTON));
      XSetForeground(dpy, gc, COLOR(buttons[i].lc));

      XTextExtents(finfo, buttons[i].label, strlen(buttons[i].label),
		   &dir, &fa, &fd, &xchar);
      x = (buttons[i].w - xchar.width) / 2;
      y = (1 + buttons[i].h -
          (xchar.ascent + xchar.descent)) / 2 + xchar.ascent + 1;
      XDrawImageString(dpy, buttons[i].map, gc, x, y,
                       buttons[i].label, strlen(buttons[i].label));

      XSetBackground(dpy, gc, COLOR(BLACK));

    } else if (buttons[i].lw != 0) {

      /*
       * draw pixmap centered in button
       */
      XSetBackground(dpy, gc, COLOR(BUTTON));
      XSetForeground(dpy, gc, COLOR(buttons[i].lc));

      pix = XCreateBitmapFromData(dpy, buttons[i].xwin, buttons[i].lb,
				  buttons[i].lw, buttons[i].lh);

      x = (1 + buttons[i].w - buttons[i].lw) / 2;
      y = (1 + buttons[i].h - buttons[i].lh) / 2 + 1;

      XCopyPlane(dpy, pix, buttons[i].map, gc, 0, 0, buttons[i].lw,
		 buttons[i].lh, x, y, 1);

      XFreePixmap(dpy, pix);

      XSetBackground(dpy, gc, COLOR(BLACK));

    }

    /*
     * draw edge of button
     */
    XSetForeground(dpy, gc, COLOR(BUT_TOP));

    XDrawLine(dpy, buttons[i].map, gc, 1, buttons[i].h - 2, 1, 1);
    XDrawLine(dpy, buttons[i].map, gc, 2, buttons[i].h - 3, 2, 2);
    XDrawLine(dpy, buttons[i].map, gc, 3, buttons[i].h - 4, 3, 3);

    XDrawLine(dpy, buttons[i].map, gc, 1, 1, buttons[i].w - 2, 1);
    XDrawLine(dpy, buttons[i].map, gc, 2, 2, buttons[i].w - 3, 2);
    XDrawLine(dpy, buttons[i].map, gc, 3, 3, buttons[i].w - 4, 3);
    XDrawLine(dpy, buttons[i].map, gc, 4, 4, buttons[i].w - 5, 4);

    XDrawPoint(dpy, buttons[i].map, gc, 4, 5);

    XSetForeground(dpy, gc, COLOR(BUT_BOT));

    XDrawLine(dpy, buttons[i].map, gc, 3, buttons[i].h - 2,
              buttons[i].w - 2, buttons[i].h - 2);
    XDrawLine(dpy, buttons[i].map, gc, 4, buttons[i].h - 3,
              buttons[i].w - 3, buttons[i].h - 3);

    XDrawLine(dpy, buttons[i].map, gc, buttons[i].w - 2,
              buttons[i].h - 2, buttons[i].w - 2, 3);
    XDrawLine(dpy, buttons[i].map, gc, buttons[i].w - 3,
              buttons[i].h - 3, buttons[i].w - 3, 4);
    XDrawLine(dpy, buttons[i].map, gc, buttons[i].w - 4,
              buttons[i].h - 4, buttons[i].w - 4, 5);

    XDrawPoint(dpy, buttons[i].map, gc, buttons[i].w - 5, buttons[i].h - 4);

    /*
     * draw frame around button
     */
    XSetForeground(dpy, gc, COLOR(BLACK));

    XDrawLine(dpy, buttons[i].map, gc, 0, buttons[i].h - 3, 0, 2);
    XDrawLine(dpy, buttons[i].map, gc, 2, 0, buttons[i].w - 3, 0);
    XDrawLine(dpy, buttons[i].map, gc, 2, buttons[i].h - 1,
              buttons[i].w - 3, buttons[i].h - 1);
    XDrawLine(dpy, buttons[i].map, gc, buttons[i].w - 1,
              buttons[i].h - 3, buttons[i].w - 1, 2);

    XDrawPoint(dpy, buttons[i].map, gc, 1, 1);
    XDrawPoint(dpy, buttons[i].map, gc, 1, buttons[i].h - 2);
    XDrawPoint(dpy, buttons[i].map, gc, buttons[i].w - 2, 1);
    XDrawPoint(dpy, buttons[i].map, gc, buttons[i].w - 2, buttons[i].h - 2);

    /*
     * draw the depressed button
     */
    buttons[i].down = XCreatePixmap(dpy, buttons[i].xwin, buttons[i].w,
				    buttons[i].h, depth);

    XSetForeground(dpy, gc, pixel);
    XFillRectangle(dpy, buttons[i].down, gc, 0, 0, buttons[i].w,
                   buttons[i].h);

    XSetForeground(dpy, gc, COLOR(BUTTON));
    XFillRectangle(dpy, buttons[i].down, gc, 1, 1, buttons[i].w - 2,
                   buttons[i].h - 2);

    if (buttons[i].label != (char *)0) {

      /*
       * set small or big font in gc
       */
      switch (buttons[i].font_size) {
        case 0:
          finfo = f_small;
          break;
        case 1:
          finfo = f_big;
          break;
        case 2:
          finfo = f_med;
          break;
        default:
          finfo = f_small;
          break;
      }
      val.font = finfo->fid;
      gc_mask = GCFont;
      XChangeGC(dpy, gc, gc_mask, &val);

      /*
       * draw string centered in button
       */
      XSetBackground(dpy, gc, COLOR(BUTTON));
      XSetForeground(dpy, gc, COLOR(buttons[i].lc));

      XTextExtents(finfo, buttons[i].label, strlen(buttons[i].label),
		   &dir, &fa, &fd, &xchar);
      x = (buttons[i].w - xchar.width) / 2;
      y = (1 + buttons[i].h -
          (xchar.ascent + xchar.descent)) / 2 + xchar.ascent;
      XDrawImageString(dpy, buttons[i].down, gc, x, y,
                       buttons[i].label, strlen(buttons[i].label));

      XSetBackground(dpy, gc, COLOR(BLACK));

    } else {

      /*
       * draw pixmap centered in button
       */
      XSetBackground(dpy, gc, COLOR(BUTTON));
      XSetForeground(dpy, gc, COLOR(buttons[i].lc));

      pix = XCreateBitmapFromData(dpy, buttons[i].xwin, buttons[i].lb,
				  buttons[i].lw, buttons[i].lh);

      x = (1 + buttons[i].w - buttons[i].lw) / 2;
      y = (1 + buttons[i].h - buttons[i].lh) / 2;

      XCopyPlane(dpy, pix, buttons[i].down, gc, 0, 0, buttons[i].lw,
		 buttons[i].lh, x, y, 1);

      XFreePixmap(dpy, pix);

      XSetBackground(dpy, gc, COLOR(BLACK));

    }

    /*
     * draw edge of button
     */
    XSetForeground(dpy, gc, COLOR(BUT_TOP));

    XDrawLine(dpy, buttons[i].down, gc, 2, buttons[i].h - 4, 2, 2);
    XDrawLine(dpy, buttons[i].down, gc, 3, buttons[i].h - 5, 3, 3);

    XDrawLine(dpy, buttons[i].down, gc, 2, 2, buttons[i].w - 4, 2);
    XDrawLine(dpy, buttons[i].down, gc, 3, 3, buttons[i].w - 5, 3);

    XDrawPoint(dpy, buttons[i].down, gc, 4, 4);

    XSetForeground(dpy, gc, COLOR(BUT_BOT));

    XDrawLine(dpy, buttons[i].down, gc, 3, buttons[i].h - 3,
              buttons[i].w - 3, buttons[i].h - 3);
    XDrawLine(dpy, buttons[i].down, gc, 4, buttons[i].h - 4,
              buttons[i].w - 4, buttons[i].h - 4);

    XDrawLine(dpy, buttons[i].down, gc, buttons[i].w - 3,
              buttons[i].h - 3, buttons[i].w - 3, 3);
    XDrawLine(dpy, buttons[i].down, gc, buttons[i].w - 4,
              buttons[i].h - 4, buttons[i].w - 4, 4);

    XDrawPoint(dpy, buttons[i].down, gc, buttons[i].w - 5, buttons[i].h - 5);

    /*
     * draw frame around button
     */
    XSetForeground(dpy, gc, COLOR(BLACK));

    XDrawLine(dpy, buttons[i].down, gc, 0, buttons[i].h - 3, 0, 2);
    XDrawLine(dpy, buttons[i].down, gc, 2, 0, buttons[i].w - 3, 0);
    XDrawLine(dpy, buttons[i].down, gc, 2, buttons[i].h - 1,
              buttons[i].w - 3, buttons[i].h - 1);
    XDrawLine(dpy, buttons[i].down, gc, buttons[i].w - 1,
              buttons[i].h - 3, buttons[i].w - 1, 2);

    XDrawPoint(dpy, buttons[i].down, gc, 1, 1);
    XDrawPoint(dpy, buttons[i].down, gc, 1, buttons[i].h - 2);
    XDrawPoint(dpy, buttons[i].down, gc, buttons[i].w - 2, 1);
    XDrawPoint(dpy, buttons[i].down, gc, buttons[i].w - 2, buttons[i].h - 2);

    XDrawRectangle(dpy, buttons[i].down, gc, 1, 1, buttons[i].w - 3,
                   buttons[i].h - 3);

    XDrawPoint(dpy, buttons[i].down, gc, 2, 2);
    XDrawPoint(dpy, buttons[i].down, gc, 2, buttons[i].h - 3);
    XDrawPoint(dpy, buttons[i].down, gc, buttons[i].w - 3, 2);
    XDrawPoint(dpy, buttons[i].down, gc, buttons[i].w - 3, buttons[i].h - 3);

  }

  XFreeFont(dpy, f_big);
  XFreeFont(dpy, f_med);
  XFreeFont(dpy, f_small);

  return 0;
}

int
#ifdef __FunctionProto__
DrawButtons(void)
#else
DrawButtons()
#endif
{
  int i;

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {
    if (buttons[i].pressed) {
      XCopyArea(dpy, buttons[i].down, buttons[i].xwin, gc, 0, 0, buttons[i].w,
                buttons[i].h, 0, 0);
    } else {
      XCopyArea(dpy, buttons[i].map, buttons[i].xwin, gc, 0, 0, buttons[i].w,
                buttons[i].h, 0, 0);
    }
  }
  return 0;
}

int
#ifdef __FunctionProto__
DrawButton(int i)
#else
DrawButton(i)
int i;
#endif
{
  if (buttons[i].pressed) {
    XCopyArea(dpy, buttons[i].down, buttons[i].xwin, gc, 0, 0, buttons[i].w,
              buttons[i].h, 0, 0);
  } else {
    XCopyArea(dpy, buttons[i].map, buttons[i].xwin, gc, 0, 0, buttons[i].w,
              buttons[i].h, 0, 0);
  }
  return 0;
}

int
#ifdef __FunctionProto__
CreateKeypad(int w, int h)
#else
CreateKeypad(w, h)
int w;
int h;
#endif
{
  int i, x, y;
  int wl, wr, ws;
  XFontStruct *finfo;
  XGCValues val;
  unsigned long gc_mask;
  Pixmap pix;
  unsigned long pixel;
  int cut;

  /*
   * draw the nice labels around the buttons
   */
  keypad.width = w;
  keypad.height = h;

  keypad.pixmap = XCreatePixmap(dpy, mainW, w, h, depth);

  XSetBackground(dpy, gc, COLOR(PAD));
  XSetForeground(dpy, gc, COLOR(PAD));

  XFillRectangle(dpy, keypad.pixmap, gc, 0, 0, w, h);

  XSetBackground(dpy, gc, COLOR(DISP_PAD));
  XSetForeground(dpy, gc, COLOR(DISP_PAD));

  cut = KEYBOARD_OFFSET_Y + buttons[BUTTON_MTH].y - 19;

  XFillRectangle(dpy, keypad.pixmap, gc, 0, 0, w, cut);

  /*
   * draw the character labels
   */
  finfo = XLoadQueryFont(dpy, "6x12");
  val.font = finfo->fid;
  gc_mask = GCFont;
  XChangeGC(dpy, gc, gc_mask, &val);

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {

    if (i < BUTTON_MTH)
      pixel = COLOR(DISP_PAD);
    else
      pixel = COLOR(PAD);

    if (buttons[i].letter != (char *)0) {

      XSetBackground(dpy, gc, pixel);
      XSetForeground(dpy, gc, COLOR(WHITE));

      x = KEYBOARD_OFFSET_X + buttons[i].x + buttons[i].w -
          SmallTextWidth(finfo, buttons[i].letter, 1) / 2 + 5;
      y = KEYBOARD_OFFSET_Y + buttons[i].y + buttons[i].h - 2;

      DrawSmallString(dpy, keypad.pixmap, gc, finfo, x, y,
                      buttons[i].letter, 1);
    }
  }

  XFreeFont(dpy, finfo);

  /*
   * draw the bottom labels
   */
  finfo = XLoadQueryFont(dpy, "6x12");
  val.font = finfo->fid;
  gc_mask = GCFont;
  XChangeGC(dpy, gc, gc_mask, &val);

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {

    if (buttons[i].sub != (char *)0) {

      XSetBackground(dpy, gc, pixel);
      XSetForeground(dpy, gc, COLOR(WHITE));

      x = KEYBOARD_OFFSET_X + buttons[i].x + (1 + buttons[i].w -
          SmallTextWidth(finfo, buttons[i].sub, strlen(buttons[i].sub))) / 2;
      y = KEYBOARD_OFFSET_Y + buttons[i].y + buttons[i].h +
          finfo->ascent + 2;

      DrawSmallString(dpy, keypad.pixmap, gc, finfo, x, y,
                      buttons[i].sub, strlen(buttons[i].sub));
    }
  }

  XFreeFont(dpy, finfo);

  /*
   * draw the left labels
   */
  finfo = XLoadQueryFont(dpy, "6x12");
  val.font = finfo->fid;
  gc_mask = GCFont;
  XChangeGC(dpy, gc, gc_mask, &val);

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {

    if (buttons[i].left != (char *)0) {

      if (buttons[i].is_menu) {

        /*
         * draw the dark shade under the label
         */
	pix = XCreatePixmap(dpy, keypad.pixmap, 44, 9, depth);

        XSetForeground(dpy, gc, COLOR(BLACK));

	XFillRectangle(dpy, pix, gc, 0, 0, 44, 9);

        XSetBackground(dpy, gc, COLOR(BLACK));
	XSetForeground(dpy, gc, COLOR(LEFT));

	x = (45 - SmallTextWidth(finfo, buttons[i].left,
                                 strlen(buttons[i].left))) / 2;

        DrawSmallString(dpy, pix, gc, finfo, x, 8,
                        buttons[i].left, strlen(buttons[i].left));

        XSetForeground(dpy, gc, pixel);

	XDrawPoint(dpy, pix, gc, 0, 0); 
	XDrawPoint(dpy, pix, gc, 0, 8); 
	XDrawPoint(dpy, pix, gc, 43, 0); 
	XDrawPoint(dpy, pix, gc, 43, 8); 

        x = KEYBOARD_OFFSET_X + buttons[i].x + (buttons[i].w - 44) / 2;
        y = KEYBOARD_OFFSET_Y + buttons[i].y - finfo->ascent - finfo->descent;

        XCopyArea(dpy, pix, keypad.pixmap, gc, 0, 0, 44, 9, x, y);

	XFreePixmap(dpy, pix);

      } else {

        XSetBackground(dpy, gc, pixel);
        XSetForeground(dpy, gc, COLOR(LEFT));

        if (buttons[i].right == (char *)0) {	/* centered label */

          x = KEYBOARD_OFFSET_X + buttons[i].x + (1 + buttons[i].w -
              SmallTextWidth(finfo, buttons[i].left,
                             strlen(buttons[i].left))) / 2;

        } else {					/* label to the left */

	  wl = SmallTextWidth(finfo, buttons[i].left, strlen(buttons[i].left));
	  wr = SmallTextWidth(finfo, buttons[i].right, strlen(buttons[i].right));
	  ws = SmallTextWidth(finfo, " ", 1);

          x = KEYBOARD_OFFSET_X + buttons[i].x + (1 + buttons[i].w -
              (wl + wr + ws)) / 2;

        }

        y = KEYBOARD_OFFSET_Y + buttons[i].y - finfo->descent;

        DrawSmallString(dpy, keypad.pixmap, gc, finfo, x, y,
                        buttons[i].left, strlen(buttons[i].left));
      }
    }
  }

  XFreeFont(dpy, finfo);

  /*
   * draw the right labels
   */
  finfo = XLoadQueryFont(dpy, "6x12");
  val.font = finfo->fid;
  gc_mask = GCFont;
  XChangeGC(dpy, gc, gc_mask, &val);

  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {

    if (i < BUTTON_MTH)
      pixel = COLOR(DISP_PAD);
    else
      pixel = COLOR(PAD);

    if (buttons[i].right != (char *)0) {

      XSetBackground(dpy, gc, pixel);
      XSetForeground(dpy, gc, COLOR(RIGHT));

      if (buttons[i].left == (char *)0) {	/* centered label */

        x = KEYBOARD_OFFSET_X + buttons[i].x + (1 + buttons[i].w -
            SmallTextWidth(finfo, buttons[i].left,
                           strlen(buttons[i].left))) / 2;

      } else {					/* label to the right */

	wl = SmallTextWidth(finfo, buttons[i].left, strlen(buttons[i].left));
	wr = SmallTextWidth(finfo, buttons[i].right, strlen(buttons[i].right));
	ws = SmallTextWidth(finfo, " ", 1);

        x = KEYBOARD_OFFSET_X + buttons[i].x + (1 + buttons[i].w -
            (wl + wr + ws)) / 2 + wl + ws;

      }

      y = KEYBOARD_OFFSET_Y + buttons[i].y - finfo->descent;

      DrawSmallString(dpy, keypad.pixmap, gc, finfo, x, y,
                      buttons[i].right, strlen(buttons[i].right));
    }
  }

  XFreeFont(dpy, finfo);

  /*
   * at last draw the --- LAST --- thing
   */

  XSetBackground(dpy, gc, COLOR(PAD));
  XSetForeground(dpy, gc, COLOR(WHITE));

  pix = XCreateBitmapFromData(dpy, keypad.pixmap, last_bits,
			      last_width, last_height);

  x = KEYBOARD_OFFSET_X + buttons[BUTTON_1].x + buttons[BUTTON_1].w +
      (buttons[BUTTON_2].x - buttons[BUTTON_1].x - buttons[BUTTON_1].w) / 2;
  y = KEYBOARD_OFFSET_Y + buttons[BUTTON_5].y + buttons[BUTTON_5].h + 2;

  XCopyPlane(dpy, pix, keypad.pixmap, gc, 0, 0, last_width, last_height,
	     x, y, 1);

  XFreePixmap(dpy, pix);

  /*
   * draw the frame around the display
   */

  XSetForeground(dpy, gc, COLOR(DISP_PAD_TOP));

  for (i = 0; i < DISP_FRAME; i++) {
    XDrawLine(dpy, keypad.pixmap, gc,
              DISPLAY_OFFSET_X - i,
              DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i,
              DISPLAY_OFFSET_X + DISPLAY_WIDTH + i,
              DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i);
    XDrawLine(dpy, keypad.pixmap, gc,
              DISPLAY_OFFSET_X - i,
              DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i + 1,
              DISPLAY_OFFSET_X + DISPLAY_WIDTH + i,
              DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i + 1);
    XDrawLine(dpy, keypad.pixmap, gc,
	      DISPLAY_OFFSET_X + DISPLAY_WIDTH + i,
	      DISPLAY_OFFSET_Y - i,
	      DISPLAY_OFFSET_X + DISPLAY_WIDTH + i,
	      DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i);
  }

  XSetForeground(dpy, gc, COLOR(DISP_PAD_BOT));

  for (i = 0; i < DISP_FRAME; i++) {
    XDrawLine(dpy, keypad.pixmap, gc,
              DISPLAY_OFFSET_X - i - 1,
              DISPLAY_OFFSET_Y - i - 1,
              DISPLAY_OFFSET_X + DISPLAY_WIDTH + i - 1,
              DISPLAY_OFFSET_Y - i - 1);
    XDrawLine(dpy, keypad.pixmap, gc,
	      DISPLAY_OFFSET_X - i - 1,
	      DISPLAY_OFFSET_Y - i - 1,
	      DISPLAY_OFFSET_X - i - 1,
	      DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * i - 1);
  }

  /*
   * round off corners
   */
  XSetForeground(dpy, gc, COLOR(DISP_PAD));

  XDrawLine(dpy, keypad.pixmap, gc,
            DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y - DISP_FRAME,
	    DISPLAY_OFFSET_X - DISP_FRAME + 3,
	    DISPLAY_OFFSET_Y - DISP_FRAME);
  XDrawLine(dpy, keypad.pixmap, gc,
            DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y - DISP_FRAME,
	    DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y - DISP_FRAME + 3);
  XDrawPoint(dpy, keypad.pixmap, gc,
	     DISPLAY_OFFSET_X - DISP_FRAME + 1,
	     DISPLAY_OFFSET_Y - DISP_FRAME + 1);

  XDrawLine(dpy, keypad.pixmap, gc,
            DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 4,
	    DISPLAY_OFFSET_Y - DISP_FRAME,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y - DISP_FRAME);
  XDrawLine(dpy, keypad.pixmap, gc,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y - DISP_FRAME,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y - DISP_FRAME + 3);
  XDrawPoint(dpy, keypad.pixmap, gc,
	     DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 2,
	     DISPLAY_OFFSET_Y - DISP_FRAME + 1);

  XDrawLine(dpy, keypad.pixmap, gc,
            DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 4,
	    DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1);
  XDrawLine(dpy, keypad.pixmap, gc,
            DISPLAY_OFFSET_X - DISP_FRAME,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1,
	    DISPLAY_OFFSET_X - DISP_FRAME + 3,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1);
  XDrawPoint(dpy, keypad.pixmap, gc,
	     DISPLAY_OFFSET_X - DISP_FRAME + 1,
	     DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 2);

  XDrawLine(dpy, keypad.pixmap, gc,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 4,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1);
  XDrawLine(dpy, keypad.pixmap, gc,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 4,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 1,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 1);
  XDrawPoint(dpy, keypad.pixmap, gc,
	     DISPLAY_OFFSET_X + DISPLAY_WIDTH + DISP_FRAME - 2,
	     DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + 2 * DISP_FRAME - 2);

  /*
   * simulate rounded lcd corners
   */
  XSetForeground(dpy, gc, COLOR(LCD));

  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X - 1,
	    DISPLAY_OFFSET_Y + 1, DISPLAY_OFFSET_X - 1,
            DISPLAY_OFFSET_Y + DISPLAY_HEIGHT - 2);
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X + 1,
	    DISPLAY_OFFSET_Y - 1, DISPLAY_OFFSET_X + DISPLAY_WIDTH - 2,
            DISPLAY_OFFSET_Y - 1); 
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X + 1,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT,
	    DISPLAY_OFFSET_X + DISPLAY_WIDTH - 2,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT);
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X + DISPLAY_WIDTH,
	    DISPLAY_OFFSET_Y + 1, DISPLAY_OFFSET_X + DISPLAY_WIDTH,
	    DISPLAY_OFFSET_Y + DISPLAY_HEIGHT - 2);

  /*
   * lower the whole thing
   */
  XSetForeground(dpy, gc, COLOR(PAD_TOP));

  /* bottom lines */
  XDrawLine(dpy, keypad.pixmap, gc, 1, keypad.height - 1,
	    keypad.width - 1, keypad.height - 1);
  XDrawLine(dpy, keypad.pixmap, gc, 2, keypad.height - 2,
	    keypad.width - 2, keypad.height - 2);

  /* right lines */
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 1,
	    keypad.height - 1, keypad.width - 1, cut);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 2,
	    keypad.height - 2, keypad.width - 2, cut);

  XSetForeground(dpy, gc, COLOR(DISP_PAD_TOP));

  /* right lines */
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 1,
	    cut - 1, keypad.width - 1, 1);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 2,
	    cut - 1, keypad.width - 2, 2);

  XSetForeground(dpy, gc, COLOR(DISP_PAD_BOT));

  /* top lines */
  XDrawLine(dpy, keypad.pixmap, gc, 0, 0, keypad.width - 2, 0);
  XDrawLine(dpy, keypad.pixmap, gc, 1, 1, keypad.width - 3, 1);

  /* left lines */
  XDrawLine(dpy, keypad.pixmap, gc, 0, cut - 1, 0, 0);
  XDrawLine(dpy, keypad.pixmap, gc, 1, cut - 1, 1, 1);

  XSetForeground(dpy, gc, COLOR(PAD_BOT));

  /* left lines */
  XDrawLine(dpy, keypad.pixmap, gc, 0, keypad.height - 2, 0, cut);
  XDrawLine(dpy, keypad.pixmap, gc, 1, keypad.height - 3, 1, cut);

  /*
   * lower the menu buttons
   */
  XSetForeground(dpy, gc, COLOR(PAD_TOP));

  /* bottom lines */
  XDrawLine(dpy, keypad.pixmap, gc, 3, keypad.height - 3,
	    keypad.width - 3, keypad.height - 3);
  XDrawLine(dpy, keypad.pixmap, gc, 4, keypad.height - 4,
	    keypad.width - 4, keypad.height - 4);

  /* right lines */
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 3,
	    keypad.height - 3, keypad.width - 3, cut);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 4,
	    keypad.height - 4, keypad.width - 4, cut);

  XSetForeground(dpy, gc, COLOR(DISP_PAD_TOP));

  /* right lines */
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 3,
	    cut - 1, keypad.width - 3, KEYBOARD_OFFSET_Y - 24);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 4,
	    cut - 1, keypad.width - 4, KEYBOARD_OFFSET_Y - 23);

  XSetForeground(dpy, gc, COLOR(DISP_PAD_BOT));

  /* top lines */
  XDrawLine(dpy, keypad.pixmap, gc, 2, KEYBOARD_OFFSET_Y - 25,
            keypad.width - 4, KEYBOARD_OFFSET_Y - 25);
  XDrawLine(dpy, keypad.pixmap, gc, 3, KEYBOARD_OFFSET_Y - 24,
            keypad.width - 5, KEYBOARD_OFFSET_Y - 24);

  /* left lines */
  XDrawLine(dpy, keypad.pixmap, gc, 2, cut - 1, 2, KEYBOARD_OFFSET_Y - 24);
  XDrawLine(dpy, keypad.pixmap, gc, 3, cut - 1, 3, KEYBOARD_OFFSET_Y - 23);

  XSetForeground(dpy, gc, COLOR(PAD_BOT));

  /* left lines */
  XDrawLine(dpy, keypad.pixmap, gc, 2, keypad.height - 4, 2, cut);
  XDrawLine(dpy, keypad.pixmap, gc, 3, keypad.height - 5, 3, cut);

  /*
   * lower the keyboard
   */
  XSetForeground(dpy, gc, COLOR(PAD_TOP));

  /* bottom lines */
  XDrawLine(dpy, keypad.pixmap, gc, 5, keypad.height - 5,
	    keypad.width - 3, keypad.height - 5);
  XDrawLine(dpy, keypad.pixmap, gc, 6, keypad.height - 6,
	    keypad.width - 4, keypad.height - 6);

  /* right lines */
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 5,
	    keypad.height - 5, keypad.width - 5, cut + 1);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 6,
	    keypad.height - 6, keypad.width - 6, cut + 2);

  XSetForeground(dpy, gc, COLOR(DISP_PAD_BOT));

  /* top lines */
  XDrawLine(dpy, keypad.pixmap, gc, 4, cut,
            keypad.width - 6, cut);
  XDrawLine(dpy, keypad.pixmap, gc, 5, cut + 1,
            keypad.width - 7, cut + 1);

  XSetForeground(dpy, gc, COLOR(PAD_BOT));

  /* left lines */
  XDrawLine(dpy, keypad.pixmap, gc, 4, keypad.height - 6, 4, cut + 1);
  XDrawLine(dpy, keypad.pixmap, gc, 5, keypad.height - 7, 5, cut + 2);

  /*
   * round off the bottom edge
   */
  XSetForeground(dpy, gc, COLOR(PAD_TOP));

  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 7, keypad.height - 7,
            keypad.width - 7, keypad.height - 14);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 8, keypad.height - 8,
            keypad.width - 8, keypad.height - 11);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 7, keypad.height - 7,
            keypad.width - 14, keypad.height - 7);
  XDrawLine(dpy, keypad.pixmap, gc, keypad.width - 7, keypad.height - 8,
            keypad.width - 11, keypad.height - 8);
  XDrawPoint(dpy, keypad.pixmap, gc, keypad.width - 9, keypad.height - 9);

  XDrawLine(dpy, keypad.pixmap, gc, 7, keypad.height - 7, 13,
            keypad.height - 7);
  XDrawLine(dpy, keypad.pixmap, gc, 8, keypad.height - 8, 10,
            keypad.height - 8);
  XSetForeground(dpy, gc, COLOR(PAD_BOT));
  XDrawLine(dpy, keypad.pixmap, gc, 6, keypad.height - 8, 6,
            keypad.height - 14);
  XDrawLine(dpy, keypad.pixmap, gc, 7, keypad.height - 9, 7,
            keypad.height - 11);

  /*
   * insert the HP Logo
   */

  XSetBackground(dpy, gc, COLOR(LOGO_BACK));
  XSetForeground(dpy, gc, COLOR(LOGO));

  pix = XCreateBitmapFromData(dpy, keypad.pixmap, hp_bits, hp_width, hp_height);

  XCopyPlane(dpy, pix, keypad.pixmap, gc, 0, 0, hp_width, hp_height,
             DISPLAY_OFFSET_X, 10, 1);

  XFreePixmap(dpy, pix);

  XSetForeground(dpy, gc, COLOR(BLACK));

  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X, 9,
            DISPLAY_OFFSET_X + hp_width - 1, 9);
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X - 1, 10,
            DISPLAY_OFFSET_X - 1, 10 + hp_height - 1);
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X, 10 + hp_height,
	    DISPLAY_OFFSET_X + hp_width - 1, 10 + hp_height);
  XDrawLine(dpy, keypad.pixmap, gc, DISPLAY_OFFSET_X + hp_width, 10,
	    DISPLAY_OFFSET_X + hp_width, 10 + hp_height - 1);

  /*
   * write the name of it
   */
  XSetBackground(dpy, gc, COLOR(DISP_PAD));
  XSetForeground(dpy, gc, COLOR(LABEL));

  x = DISPLAY_OFFSET_X;
  y = TOP_SKIP - DISP_FRAME - hp48sx_height - 3;

  pix = XCreateBitmapFromData(dpy, keypad.pixmap, hp48sx_bits,
                              hp48sx_width, hp48sx_height);

  XCopyPlane(dpy, pix, keypad.pixmap, gc, 0, 0, hp48sx_width, hp48sx_height,
             x, y, 1);

  XFreePixmap(dpy, pix);

  x = DISPLAY_OFFSET_X + DISPLAY_WIDTH - 1 - science_width;
  y = TOP_SKIP - DISP_FRAME - science_height - 4;

  pix = XCreateBitmapFromData(dpy, keypad.pixmap, science_bits,
                              science_width, science_height);

  XCopyPlane(dpy, pix, keypad.pixmap, gc, 0, 0, science_width, science_height,
             x, y, 1);

  /*
   * that's it. Ooph.
   */
  return 0;
}

int
#ifdef __FunctionProto__
DrawKeypad(void)
#else
DrawKeypad()
#endif
{
  XCopyArea(dpy, keypad.pixmap, mainW, gc, 0, 0, keypad.width,
            keypad.height, 0, 0);
  return 0;
}

#ifdef USE_SHM
int
#ifdef __FunctionProto__
handle_xerror(Display *dpy, XErrorEvent *eev)
#else
handle_xerror(dpy, eev)
Display *dpy;
XErrorEvent *eev;
#endif
{
  xerror_flag = 1;
  return 0;
}
#endif

int
#ifdef __FunctionProto__
CreateDispWindow(void)
#else
CreateDispWindow()
#endif
{
  XSetWindowAttributes xswa;
  XGCValues val;
  unsigned long gc_mask;
  static XRectangle rect;

  /*
   * create the display subwindow
   */
  disp.w = DISPLAY_WIDTH;
  disp.h = DISPLAY_HEIGHT;

  disp.win = XCreateSimpleWindow(dpy, mainW,
			DISPLAY_OFFSET_X, DISPLAY_OFFSET_Y,
			disp.w, disp.h, 0,
			BlackPixel(dpy, screen), COLOR(LCD));

  disp.mapped = 1;
  disp.display_update = UPDATE_DISP | UPDATE_MENU;

  xswa.event_mask = ExposureMask | StructureNotifyMask;
  xswa.backing_store = Always;

  XChangeWindowAttributes(dpy, disp.win, CWEventMask|CWBackingStore, &xswa);

  /*
   * set up the gc
   */
  val.foreground = COLOR(LCD);
  val.background = COLOR(LCD);
  val.function = GXcopy;
  gc_mask = GCForeground | GCBackground | GCFunction;
  disp.gc = XCreateGC(dpy, mainW, gc_mask, &val);

  XSetForeground(dpy, disp.gc, COLOR(PIXEL));

#ifdef USE_SHM

  xerror_flag = 0;
  XSetErrorHandler(handle_xerror);
  XFlush(dpy);

  disp.disp_image = NULL;
  disp.menu_image = NULL;
  if (shm_flag) {

    /*
     * create XShmImage for DISP
     */
    disp.disp_image = XShmCreateImage(dpy, None, 1, XYBitmap, NULL,
                                      &disp.disp_info, 288, 128);
    if (disp.disp_image == NULL) {
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in CreateImage(DISP), disabling.\n");
#endif
      goto shm_error;
    }

    /*
     * get ID of shared memory block for DISP
     */
    disp.disp_info.shmid = shmget(IPC_PRIVATE,
            (disp.disp_image->bytes_per_line * disp.disp_image->height),
            IPC_CREAT | 0777);
    if (disp.disp_info.shmid < 0) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in shmget(DISP), disabling.\n");
#endif
      goto shm_error;
    }

    /*
     * get address of shared memory block for DISP
     */
    disp.disp_info.shmaddr = (char *)shmat(disp.disp_info.shmid, 0, 0);
    if (disp.disp_info.shmaddr == ((char *)-1)) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in shmget(DISP), disabling.\n");
#endif
      goto shm_error;
    }
    disp.disp_image->data = disp.disp_info.shmaddr;
    disp.disp_info.readOnly = False;
    XShmAttach(dpy, &disp.disp_info);

    /*
     * create XShmImage for MENU
     */
    disp.menu_image = XShmCreateImage(dpy, None, 1, XYBitmap, NULL,
                                      &disp.menu_info, 272, 16);
    if (disp.menu_image == NULL) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in CreateImage(MENU), disabling.\n");
#endif
      goto shm_error;
    }

    /*
     * get ID of shared memory block for MENU
     */
    disp.menu_info.shmid = shmget(IPC_PRIVATE,
            (disp.menu_image->bytes_per_line * disp.menu_image->height),
            IPC_CREAT | 0777);
    if (disp.menu_info.shmid < 0) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      XDestroyImage(disp.menu_image);
      disp.menu_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in shmget(MENU), disabling.\n");
#endif
      goto shm_error;
    }

    /*
     * get address of shared memory block for MENU
     */
    disp.menu_info.shmaddr = (char *)shmat(disp.menu_info.shmid, 0, 0);
    if (disp.menu_info.shmaddr == ((char *)-1)) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      XDestroyImage(disp.menu_image);
      disp.menu_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in shmget(MENU), disabling.\n");
#endif
      goto shm_error;
    }
    disp.menu_image->data = disp.menu_info.shmaddr;
    disp.menu_info.readOnly = False;
    XShmAttach(dpy, &disp.menu_info);

    XSync(dpy, 0);

    if (xerror_flag) {
      XDestroyImage(disp.disp_image);
      disp.disp_image = NULL;
      XDestroyImage(disp.menu_image);
      disp.menu_image = NULL;
      shm_flag = 0;
#ifdef DEBUG_SHM
      fprintf(stderr, "XShm error in shmget(MENU), disabling.\n");
#endif
      goto shm_error;
    } else {
      shmctl(disp.disp_info.shmid, IPC_RMID, 0);
      shmctl(disp.menu_info.shmid, IPC_RMID, 0);
    }

    memset(disp.disp_image->data, 0,
           (disp.disp_image->bytes_per_line * disp.disp_image->height));
    memset(disp.menu_image->data, 0,
           (disp.menu_image->bytes_per_line * disp.menu_image->height));

#ifdef DEBUG_SHM
    fprintf(stderr, "Using shared memory.\n");
#endif
  }
  shm_error:

  XSetErrorHandler(NULL);
  XFlush(dpy);

  if (!shm_flag) {
#endif
    rect.x = 5;
    rect.y = 0;
    rect.width = 262;
    rect.height = disp.h;
    XSetClipRectangles(dpy, disp.gc, 0, 0, &rect, 1, Unsorted);
#ifdef USE_SHM
  }
#endif

  return 0;
}

int
#ifdef __FunctionProto__
CreateMainWindow(void)
#else
CreateMainWindow()
#endif
{
  XSizeHints hint;
  XWMHints wmh;
  XSetWindowAttributes xswa;
  XClassHint clh;
  char *name;
  XTextProperty wname, iname;
  Visual *visual;
  Atom protocols[1];
  unsigned int mask;
  unsigned int class;
  XGCValues val;
  unsigned long gc_mask;
  int width, height;
  Pixmap icon_pix, icon_mask;

  class = InputOutput;
  depth = DefaultDepth(dpy, screen);
  visual = DefaultVisual(dpy, screen);
  cmap = DefaultColormap(dpy, screen);

  /*
   * allocate my colors
   */
  AllocColors();

  xswa.background_pixel = COLOR(PAD);
  xswa.border_pixel = WhitePixel(dpy, screen);
  xswa.backing_store = Always;
  xswa.event_mask = KeyPressMask | KeyReleaseMask | ButtonPressMask |
                    ButtonReleaseMask | ExposureMask | FocusChangeMask |
                    EnterWindowMask | StructureNotifyMask;
  xswa.colormap = cmap;
  mask = CWBackPixel | CWBorderPixel | CWBackingStore | CWEventMask |
         CWColormap;

  name = (char *)malloc(128);
  if (name == (char *)0) {
    fprintf(stderr, "Can\'t malloc 128 bytes, exit\n");
    exit (1);
  }
  sprintf(name, "%s-%d.%d.%d", progname, saturn.version[0],
          saturn.version[1], saturn.version[2]);

  clh.res_class = clh.res_name = name;

  if(!XStringListToTextProperty(&name, 1, &wname))
    return -1;
  if(!XStringListToTextProperty(&name, 1, &iname))
    return -1;

  /*
   * create the window
   */
  width = KEYBOARD_WIDTH + 2 * SIDE_SKIP;
  height = DISPLAY_OFFSET_Y + DISPLAY_HEIGHT + DISP_KBD_SKIP +
           KEYBOARD_HEIGHT + BOTTOM_SKIP;

  mainW = XCreateWindow(dpy, RootWindow(dpy, screen), 0, 0,
                        width, height, 1, depth, class,
                        visual, mask, &xswa);

  if (mainW == 0)
    return -1;

  /*
   * should parse -geometry ...
   */
  hint.x = hint.y = 0;
  hint.min_width = hint.max_width = hint.width = width;
  hint.min_height = hint.max_height = hint.height = height;
  hint.flags = PSize | PPosition | PMinSize | PMaxSize;

  wmh.initial_state = NormalState;
  wmh.input = True;
  wmh.flags = StateHint | InputHint;

  /*
   * create and set the icon
   */
  icon_pix = XCreateBitmapFromData(dpy, mainW, hp48_icon_bits,
		hp48_icon_width, hp48_icon_height);
  icon_mask = XCreateBitmapFromData(dpy, mainW, hp48_icon_mask_bits,
		hp48_icon_mask_width, hp48_icon_mask_height);
  if (icon_pix) {
    wmh.icon_pixmap = icon_pix;
    wmh.icon_mask = icon_mask;
    wmh.flags |= (IconPixmapHint | IconMaskHint);
  }

  XSetWMNormalHints(dpy, mainW, &hint);
  XSetWMHints(dpy, mainW, &wmh);
  XSetClassHint(dpy, mainW, &clh);
  XSetWMName(dpy, mainW, &wname);
  XSetWMIconName(dpy, mainW, &iname);

  /*
   * turn on WM_DELETE_WINDOW
   */
  wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", 0);
  wm_protocols = XInternAtom(dpy, "WM_PROTOCOLS", 0);
  protocols[0] = wm_delete_window;
  XSetWMProtocols(dpy, mainW, protocols, 1);

  /*
   * set up the GC's
   */
  val.foreground = COLOR(WHITE);
  val.background = COLOR(BLACK);
  val.function = GXcopy;
  gc_mask = GCForeground | GCBackground | GCFunction;
  gc = XCreateGC(dpy, mainW, gc_mask, &val);

  /*
   * create the display
   */
  CreateDispWindow();

  /*
   * create the buttons
   */
  CreateButtons();

  /*
   * create the keypad
   */
  CreateKeypad(width, height);

  /*
   * map the window
   */
  XMapWindow(dpy, mainW);
  XMapSubwindows(dpy, mainW);

  /*
   * wait for map
   */
  while (1) {
    XEvent xev;

    XNextEvent(dpy, &xev);
    if (xev.type == MapNotify && xev.xmap.event == mainW)
      break;
  }

  DrawKeypad();
  DrawButtons();

#ifdef USE_SHM
  if (shm_flag) {
    XFillRectangle(dpy, disp.win, disp.gc, 5, 20, 262, 128);
  }
#endif

  return 0;
}

int
#ifdef __FunctionProto__
button_pressed(int i, XEvent *xev)
#else
button_pressed(i, xev)
int i;
XEvent *xev;
#endif
{
  int code;
  int b, r, c;

  code = buttons[i].code;
  buttons[i].pressed = 1;
  if (code == 0x8000) {
    for (b = 0; b < 9; b++)
      saturn.keybuf.rows[b] |= 0x8000;
    do_kbd_int();
  } else {
    r = code >> 4;
    c = 1 << (code & 0xf);
    if ((saturn.keybuf.rows[r] & c) == 0) {
      if (saturn.kbd_ien) {
        do_kbd_int();
      }
      saturn.keybuf.rows[r] |= c;
    }
  }
  return 0;
}

int
#ifdef __FunctionProto__
button_released(int b, XEvent *xev)
#else
button_released(b, xev)
int b;
XEvent *xev;
#endif
{
  int i, code;

  code = buttons[i].code;
  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {
    if (buttons[i].pressed) {
      buttons[i].pressed = 0;
      DrawButton(i);
    }
  }
  memset(&saturn.keybuf, 0, sizeof(saturn.keybuf));
  return 0;
}

int
#ifdef __FunctionProto__
key_event(int b, XEvent *xev)
#else
key_event(b, xev)
int b;
XEvent *xev;
#endif
{
  int code;
  int i, r, c;

  code = buttons[b].code;
  if (xev->type == KeyPress) {
    if (code == 0x8000) {
      for (i = 0; i < 9; i++)
        saturn.keybuf.rows[i] |= 0x8000;
      do_kbd_int();
    } else {
      r = code >> 4;
      c = 1 << (code & 0xf);
      if ((saturn.keybuf.rows[r] & c) == 0) {
        if (saturn.kbd_ien) {
          do_kbd_int();
        }
        saturn.keybuf.rows[r] |= c;
      }
    }
  } else {
    for (i = BUTTON_A; i <= LAST_BUTTON; i++) {
      if (buttons[i].pressed) {
        buttons[i].pressed = 0;
        DrawButton(i);
      }
    }
    memset(&saturn.keybuf, 0, sizeof(saturn.keybuf));
  }
  return 0;
}

#ifdef USE_SHM
void
#ifdef __FunctionProto__
refresh_display(void)
#else
refresh_display()
#endif
{
  if (shm_flag) { 
    if (disp.display_update & UPDATE_DISP) {
      XShmPutImage(dpy, disp.win, disp.gc, disp.disp_image,
                   disp.offset, 0, 5, 20, 262, disp.lines + 2, 0);
    }
    if ((display.lines < 63) && (disp.display_update & UPDATE_MENU)) {
      XShmPutImage(dpy, disp.win, disp.gc, disp.menu_image,
                   0, disp.lines - 110, 5, 22 + disp.lines,
                   262, 126 - disp.lines, 0);
    }
    disp.display_update = 0;
  }
}
#endif

int
#ifdef __FunctionProto__
GetEvent(void)
#else
GetEvent()
#endif
{
  XEvent xev;
  XKeyboardControl kbd_ctrl;
  XKeyboardState kbd_state;
  XClientMessageEvent *cm;
  static int auto_repeat = -1;
  int i, wake;
  int bufs = 1;
  char buf[1];
  KeySym sym;
  XComposeStatus compose;

  wake = 0;

  while (XPending(dpy) > 0) {

    XNextEvent(dpy, &xev);

    switch (xev.type) {

      case Expose:
      case NoExpose:

	break;

      case UnmapNotify:

        disp.mapped = 0;
        break;

      case MapNotify:

        disp.mapped = 1;
        update_display();
        break;

      case FocusIn:

	/*
         * we don't want to have AutoRepeatMode on
         */
        if (auto_repeat == -1) {
          XGetKeyboardControl(dpy, &kbd_state);
          auto_repeat = kbd_state.global_auto_repeat; 
        }

        kbd_ctrl.auto_repeat_mode = AutoRepeatModeOff;
        XChangeKeyboardControl(dpy, KBAutoRepeatMode, &kbd_ctrl);

        break;

      case FocusOut:

        /*
         * restore AutoRepeatMode
         */
        if (auto_repeat != -1) {
          kbd_ctrl.auto_repeat_mode = auto_repeat;
          XChangeKeyboardControl(dpy, KBAutoRepeatMode, &kbd_ctrl);
          auto_repeat = -1;
        }

        break;

      case ButtonPress:

        if (xev.xbutton.button == Button1 || xev.xbutton.button == Button2) {
	  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {
	    if (xev.xbutton.subwindow == buttons[i].xwin) {
	      button_pressed(i, &xev);
              wake = 1;
              DrawButton(i);
              break;
            }
	  }
        }
	break;

      case ButtonRelease:

        if (xev.xbutton.button == Button1) {
	  for (i = BUTTON_A; i <= LAST_BUTTON; i++) {
	    if (xev.xbutton.subwindow == buttons[i].xwin) {
	      button_released(i, &xev);
              wake = 1;
              DrawButton(i);
              break;
            }
	  }
        }
	break;

      case EnterNotify:

	if ((xev.xcrossing.state & Button1Mask) != 0) {
          if (xev.xcrossing.mode == NotifyNormal) {
            button_released(-1, &xev);
            wake = 1;
          }
        }
        break;

      case KeyPress:
      case KeyRelease:

        buf[0] = '\0';
        XLookupString(&xev.xkey, buf, bufs, &sym, &compose);
        
        switch (sym) {
	  case XK_KP_0:
          case XK_0:
            key_event(BUTTON_0, &xev);
            wake = 1;
	    break;
	  case XK_KP_1:
          case XK_1:
            key_event(BUTTON_1, &xev);
            wake = 1;
	    break;
	  case XK_KP_2:
          case XK_2:
            key_event(BUTTON_2, &xev);
            wake = 1;
	    break;
	  case XK_KP_3:
          case XK_3:
            key_event(BUTTON_3, &xev);
            wake = 1;
	    break;
	  case XK_KP_4:
          case XK_4:
            key_event(BUTTON_4, &xev);
            wake = 1;
	    break;
	  case XK_KP_5:
          case XK_5:
            key_event(BUTTON_5, &xev);
            wake = 1;
	    break;
	  case XK_KP_6:
          case XK_6:
            key_event(BUTTON_6, &xev);
            wake = 1;
	    break;
	  case XK_KP_7:
          case XK_7:
            key_event(BUTTON_7, &xev);
            wake = 1;
	    break;
	  case XK_KP_8:
          case XK_8:
            key_event(BUTTON_8, &xev);
            wake = 1;
	    break;
	  case XK_KP_9:
          case XK_9:
            key_event(BUTTON_9, &xev);
            wake = 1;
	    break;
          case XK_KP_Add:
	  case XK_plus:
            key_event(BUTTON_PLUS, &xev);
            wake = 1;
	    break;
          case XK_KP_Subtract:
	  case XK_minus:
            key_event(BUTTON_MINUS, &xev);
            wake = 1;
	    break;
          case XK_KP_Divide:
	  case XK_slash:
            key_event(BUTTON_DIV, &xev);
            wake = 1;
	    break;
          case XK_KP_Multiply:
	  case XK_asterisk:
            key_event(BUTTON_MUL, &xev);
            wake = 1;
	    break;
          case XK_KP_Enter:
          case XK_Return:
            key_event(BUTTON_ENTER, &xev);
            wake = 1;
	    break;
          case XK_KP_Decimal:
          case XK_period:
            key_event(BUTTON_PERIOD, &xev);
            wake = 1;
	    break;
          case XK_space:
            key_event(BUTTON_SPC, &xev);
            wake = 1;
            break;
          case XK_Delete:
            key_event(BUTTON_DEL, &xev);
            wake = 1;
            break;
          case XK_BackSpace:
            key_event(BUTTON_BS, &xev);
            wake = 1;
            break;
          case XK_Escape:
            key_event(BUTTON_ON, &xev);
            wake = 1;
            break;
          case XK_Shift_L:
            key_event(BUTTON_SHL, &xev);
            wake = 1;
            break;
          case XK_Shift_R:
            key_event(BUTTON_SHR, &xev);
            wake = 1;
            break;
          case XK_Alt_L:
          case XK_Alt_R:
            key_event(BUTTON_ALPHA, &xev);
            wake = 1;
            break;
          case XK_a:
          case XK_F1:
            key_event(BUTTON_A, &xev);
            wake = 1;
            break;
          case XK_b:
          case XK_F2:
            key_event(BUTTON_B, &xev);
            wake = 1;
            break;
          case XK_c:
          case XK_F3:
            key_event(BUTTON_C, &xev);
            wake = 1;
            break;
          case XK_d:
          case XK_F4:
            key_event(BUTTON_D, &xev);
            wake = 1;
            break;
          case XK_e:
          case XK_F5:
            key_event(BUTTON_E, &xev);
            wake = 1;
            break;
          case XK_f:
          case XK_F6:
            key_event(BUTTON_F, &xev);
            wake = 1;
            break;
          case XK_g:
            key_event(BUTTON_MTH, &xev);
            wake = 1;
            break;
          case XK_h:
            key_event(BUTTON_PRG, &xev);
            wake = 1;
            break;
          case XK_i:
            key_event(BUTTON_CST, &xev);
            wake = 1;
            break;
          case XK_j:
            key_event(BUTTON_VAR, &xev);
            wake = 1;
            break;
          case XK_k:
          case XK_Up:
            key_event(BUTTON_UP, &xev);
            wake = 1;
            break;
          case XK_l:
            key_event(BUTTON_NXT, &xev);
            wake = 1;
            break;
          case XK_m:
            key_event(BUTTON_COLON, &xev);
            wake = 1;
            break;
          case XK_n:
            key_event(BUTTON_STO, &xev);
            wake = 1;
            break;
          case XK_o:
            key_event(BUTTON_EVAL, &xev);
            wake = 1;
            break;
          case XK_p:
          case XK_Left:
            key_event(BUTTON_LEFT, &xev);
            wake = 1;
            break;
          case XK_q:
          case XK_Down:
            key_event(BUTTON_DOWN, &xev);
            wake = 1;
            break;
          case XK_r:
          case XK_Right:
            key_event(BUTTON_RIGHT, &xev);
            wake = 1;
            break;
          case XK_s:
            key_event(BUTTON_SIN, &xev);
            wake = 1;
            break;
          case XK_t:
            key_event(BUTTON_COS, &xev);
            wake = 1;
            break;
          case XK_u:
            key_event(BUTTON_TAN, &xev);
            wake = 1;
            break;
          case XK_v:
            key_event(BUTTON_SQRT, &xev);
            wake = 1;
            break;
          case XK_w:
            key_event(BUTTON_POWER, &xev);
            wake = 1;
            break;
          case XK_x:
            key_event(BUTTON_INV, &xev);
            wake = 1;
            break;
          case XK_y:
            key_event(BUTTON_NEG, &xev);
            wake = 1;
            break;
          case XK_z:
            key_event(BUTTON_EEX, &xev);
            wake = 1;
            break;
          default:
            break;
        }
        break;

      case ClientMessage:

        cm = (XClientMessageEvent *)&xev;

        if (cm->message_type == wm_protocols &&
            cm->data.l[0] == wm_delete_window) {

          /*
           * Quit selected from window managers menu
           */

          if (auto_repeat != -1) {
            kbd_ctrl.auto_repeat_mode = auto_repeat;
            XChangeKeyboardControl(dpy, KBAutoRepeatMode, &kbd_ctrl);
          }

          exit_emulator();
	  XCloseDisplay(dpy);
	  exit (0);

        }
        break;

      default:

#ifdef DEBUG_XEVENT
        printf("Event: %d\n", xev.type);
#endif
        break;

    }
  }
  return wake;
}

